#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Handles modification of drawing entities and views
#

import objc
import string
import Generic.segment
import Generic.split
import Generic.transfer
import Generic.mirror
import Generic.move

from Foundation import NSMakePoint

#
# Generic code - select, make distances, etc
#
def select_mouse_move_cb(doc, np, tool):
    _p1 = tool.getLocation()
    _p2 = (_p1[0],np.y)
    _p3 = (np.x, np.y)
    _p4 = (np.x,_p1[1])
    _da = doc.getDA()
    _s1 = Generic.segment.Segment(_p1, _p2)
    _s2 = Generic.segment.Segment(_p2, _p3)
    _s3 = Generic.segment.Segment(_p3, _p4)
    _s4 = Generic.segment.Segment(_p4, _p1)
    _da = doc.getDA()
    _da.setTempObject(_s1)
    _da.setTempObject(_s2, False)
    _da.setTempObject(_s3, False)
    _da.setTempObject(_s4, False)


#
# Split
#
def split_init(doc, tool):
    doc.setPrompt("Click on the objects you want to split")
    tool.setHandler("initialize", split_init)
    tool.setHandler("left_button_press", split_first_left_button_press_cb)

def split_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(NSMakePoint(_x, _y))
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _objlist = _active_layer.mapPoint((_x, _y), _tol, None)
    if len(_objlist):
        for _obj, _pt in _objlist:
        
            if isinstance(_obj, Generic.segment.Segment):
                _p1, _p2 = _obj.getEndpoints()
                _lpt = _active_layer.find('point', _pt.x, _pt.y)
                if _lpt is None:
                    _active_layer.addObject(_pt)
                    _lpt = _pt
                _s1, _s2 = Generic.split.split_segment(_obj, _pt)
                _image.startAction()
                try:
                    _active_layer.addObject(_s1)
                    _active_layer.addObject(_s2)
                    _active_layer.delObject(_obj)
                finally:
                    _image.endAction()

            elif isinstance(_obj, Generic.arc.Arc):
                _arc1, _arc2 = Generic.split.split_arc(_obj, _pt)
                _image.startAction()
                try:
                    _active_layer.addObject(_arc1)
                    _active_layer.addObject(_arc2)
                    _active_layer.delObject(_obj)
                finally:
                    _image.endAction()
                    
            elif isinstance(_obj, Generic.circle.Circle):
                _arc = Generic.split.split_circle(_obj, _pt)
                _image.startAction()
                try:
                    _active_layer.addObject(_arc)
                    _active_layer.delObject(_obj)
                finally:
                    _image.endAction()
                    
            elif isinstance(_obj, Generic.polyline.Polyline):
                _lpt = _active_layer.find('point', _pt.x, _pt.y)
                _image.startAction()
                try:
                    if _lpt is None:
                        _active_layer.addObject(_pt)
                        _lpt = _pt
                    Generic.split.split_polyline(_obj, _lpt)
                finally:
                    _image.endAction()
            else:
                pass
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("mouse_move", select_mouse_move_cb)
        tool.setHandler("left_button_press", split_second_left_button_press_cb)

def split_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax, True)
    if len(_objs):
        _splitable = []
        for _obj in _objs:
            if isinstance(_obj, (Generic.segment.Segment,
                                 Generic.circle.Circle,
                                 Generic.arc.Arc,
                                 Generic.polyline.Polyline)):
                _splitable.append(_obj)
        if len(_splitable):
            Generic.split.split_objects(_active_layer, _splitable)
    tool.reset()
    split_init(doc, tool)
                    




#
# Mirror
#
def mirror_init(doc, tool):
    doc.setPrompt("Click on the mirroring construction line")
    tool.setHandler("initialize", mirror_init)
    tool.setHandler("left_button_press", mirror_first_left_button_press_cb)

def mirror_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _objdict = _image.mapPoint(_x, _y, _tol, None)
    if len(_objdict):
        _active_layer = _image.getActiveLayer()
        if _active_layer in _objdict:
            for _obj, _pt in _objdict[_active_layer]:
                if isinstance(_obj, (Generic.hcline.HCLine,
                                     Generic.vcline.VCLine,
                                     Generic.acline.ACLine,
                                     Generic.cline.CLine)):
                    tool.setHandler("left_button_press", mirror_second_left_button_press_cb)
                    tool.setMirrorLine(_obj)
                    doc.setPrompt("Click on or select the objects to mirror")
                    break

def mirror_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(NSMakePoint(_x, _y))
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _objdict = _image.mapPoint(_x, _y, _tol, None)
    if len(_objdict):
        _active_layer = _image.getActiveLayer()
        if _active_layer in _objdict:
            _objs = []
            for _obj, _pt in _objdict[_active_layer]:
                _objs.append(_obj)
            _mline = tool.getMirrorLine()
            mirror_objects(doc, tool, _objs)
    else:
        _p, _flag = _image.findPoint(_x, _y, _tol)
        if _p is not None:
            _x, _y = _p.getCoords()
        tool.setLocation(_x, _y)
        tool.setHandler("mouse_move", select_mouse_move_cb)
        tool.setHandler("left_button_press", mirror_third_left_button_press_cb)

def mirror_third_left_button_press_cb(doc, event, tool):
    tool.delHandler("mouse_move")
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _p = _active_layer.find('point', _x2, _y2)
    if _p is not None:
        _x2, _y2 = _p.getCoords()
    _x1, _y1 = tool.getLocation()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax, True)
    mirror_objects(doc, tool, _objs)
        
def mirror_objects(doc, tool, objs):
    _mline = tool.getMirrorLine()
    _image = doc.getImage()
    _mobjs = Generic.mirror.mirror_objects(_image, _mline, objs)
    tool.reset()
    doc.getDA().flushTempObjects()
    mirror_init(doc, tool)

#
# Transfer
#
def transfer_init(doc, tool):
    doc.setPrompt("Click on the objects you want to transfer to the active layer")
    tool.setHandler("initialize", transfer_init)
    tool.setHandler("left_button_press", transfer_first_left_button_press_cb)

def transfer_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(NSMakePoint(_x, _y))
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _objdict = _image.mapPoint(_x, _y, _tol, None)
    if len(_objdict):
        _image.startAction()
        try:
            for _layer in _objdict:
                if _layer is not _active_layer:
                    _objs = []
                    for _obj, _pt in _objdict[_layer]:
                        _objs.append(_obj)
                    Generic.transfer.transfer_objects(_objs, _layer, _active_layer)
        finally:
            _image.endAction()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("mouse_move", select_mouse_move_cb)
        tool.setHandler("left_button_press", transfer_second_left_button_press_cb)

def transfer_second_left_button_press_cb(doc, event, tool):
    tool.delHandler("mouse_move")
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _layers = [_image.getTopLayer()]
    _objdict = {}
    while len(_layers):
        _layer = _layers.pop()
        if _layer is not _active_layer:
            if _layer.isVisible():
                _objs = _layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
                if len(_objs):
                    _objdict[_layer] = _objs
        _layers.extend(_layer.getSublayers())
    if len(_objdict):
        _image.startAction()
        try:
            for _layer in _objdict:
                if _layer is not _active_layer:
                    _objs = _objdict[_layer]
                    Generic.transfer.transfer_objects(_objs, _layer, _active_layer)
        finally:
            _image.endAction()
    else:
        doc.getDA().flushTempObjects()
    tool.reset()
    transfer_init(doc, tool)        
    
#
# Move
#

# functions used by all the move code

def move_objects(doc, objs, tool):
    _init_func = tool.getHandler("initialize")
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _dx, _dy = tool.getDistance()
    _image.startAction()
    try:
        Generic.move.move_objects(_active_layer, objs, _dx, _dy)
    finally:
        _image.endAction()
    doc.getDA().flushTempObjects()
    tool.reset()
    _init_func(doc, tool)

def make_distance(text):
    try:
        _textrad = eval(text)
    except:
        raise ValueError, "Invalid distance: " + text
    if not isinstance(_textrad, float):
        _textrad = float(_textrad)
    return _textrad

def move_select_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(NSMakePoint(_x, _y))
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _objdict = _image.mapPoint(_x, _y, _tol, None)
    if len(_objdict):
        _active_layer = _image.getActiveLayer()
        if _active_layer in _objdict:
            _objs = []
            for _obj, _pt in _objdict[_active_layer]:
                _objs.append(_obj)
            move_objects(doc, _objs, tool)
    else:
        _pt, _flag = _image.findPoint(_x, _y, _tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
        tool.setLocation(_x, _y)
        tool.setHandler("mouse_move", select_mouse_move_cb)
        tool.setHandler("left_button_press", move_select_second_left_button_press_cb)
        
def move_select_second_left_button_press_cb(doc, event, tool):
    tool.delHandler("mouse_move")
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    _x1, _y1 = tool.getLocation()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax, True)
    move_objects(doc, _objs, tool)



# other functions


def move_horizontal_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the horizontal distance")
    tool.setHandler("initialize", move_horizontal_init)
    tool.setHandler("entry_event", move_horizontal_entry_event_cb)
    tool.setHandler("left_button_press", move_horizontal_first_left_button_press_cb)

def move_horizontal_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _flag = _image.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    tool.setHandler("left_button_press", move_horizontal_second_left_button_press_cb)
    doc.setPrompt("Click another point to define the distance")
    
def move_horizontal_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _flag = _image.findPoint(_x2, _y2, _tol)
    if _pt is not None:
        _x2, _y2 = _pt.getCoords()
    _x1, _y1 = tool.getLocation()
    tool.setDistance((_x2 - _x1), 0.0)
    tool.clearLocation()
    tool.delHandler("entry_event")
    tool.setHandler("left_button_press", move_select_first_left_button_press_cb)
    doc.setPrompt("Click on or select the objects to move")
    
def move_horizontal_entry_event_cb(doc, text, tool):
    if len(text):
        _dist = make_distance(text)
        tool.setDistance(_dist, 0.0)
        doc.setPrompt("Click on or select the objects to move")
        tool.delHandler("entry_event")
        tool.setHandler("left_button_press", move_select_first_left_button_press_cb)

def move_vertical_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the vertical distance")
    tool.setHandler("initialize", move_vertical_init)
    tool.setHandler("entry_event", move_vertical_entry_event_cb)
    tool.setHandler("left_button_press", move_vertical_first_left_button_press_cb)

def move_vertical_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _flag = _image.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    tool.setHandler("left_button_press", move_vertical_second_left_button_press_cb)
    doc.setPrompt("Click another point to define the distance")
    
def move_vertical_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _flag = _image.findPoint(_x2, _y2, _tol)
    if _pt is not None:
        _x2, _y2 = _pt.getCoords()
    _x1, _y1 = tool.getLocation()
    tool.setDistance(0.0, (_y2 - _y1))
    tool.clearLocation()
    tool.delHandler("entry_event")
    tool.setHandler("left_button_press", move_select_first_left_button_press_cb)
    doc.setPrompt("Click on or select the objects to move")

def move_vertical_entry_event_cb(doc, text, tool):
    if len(text):
        _dist = make_distance(text)
        tool.setDistance(0.0, _dist)
        doc.setPrompt("Click on or select the objects to move")
        tool.delHandler("entry_event")
        tool.setHandler("left_button_press", move_select_first_left_button_press_cb)


def move_free_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the distance to move as 'h, v'")
    tool.setHandler("initialize", move_free_init)
    tool.setHandler("left_button_press", move_free_first_left_button_press_cb)
    tool.setHandler("entry_event", move_free_entry_event_cb)

def move_free_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _flag = _image.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setLocation(_x, _y)
    tool.setHandler("left_button_press", move_free_second_left_button_press_cb)
    doc.setPrompt("Click another point to define the distance")
    
def move_free_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _flag = _image.findPoint(_x2, _y2, _tol)
    if _pt is not None:
        _x2, _y2 = _pt.getCoords()
    _x1, _y1 = tool.getLocation()
    tool.setDistance((_x2 - _x1), (_y2 - _y1))
    tool.clearLocation()
    tool.setHandler("left_button_press", move_select_first_left_button_press_cb)
    doc.setPrompt("Click on or select the objects to move")

def move_free_entry_event_cb(doc, text, tool):
    if len(text):
        _str = string.split(text, ",")
        if len(_str) != 2:
            raise ValueError, "Invalid distance (must be h, v): " + text
        _hdist = make_distance(_str[0])
        _vdist = make_distance(_str[1])
        tool.setDistance(_hdist, _vdist)
        doc.setPrompt("Click on or select the objects to move")
        tool.delHandler("entry_event")
        tool.setHandler("left_button_press", move_select_first_left_button_press_cb)


#
# Stretch
#

def stretch_select_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(NSMakePoint(_x, _y))
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _pt = _active_layer.find('point', _x, _y, _tol)
    if _pt is not None:
        _dx, _dy = tool.getDistance()
        _pt.move(_dx, _dy)
        # should do this better . . .
        _init_func = tool.getHandler("initialize")
        tool.reset()
        _init_func(doc, tool)
        doc.getDA().setNeedsDisplay_(True) 
        
    else:
        tool.setLocation(_x, _y)
        tool.setHandler("mouse_move", select_mouse_move_cb)
        tool.setHandler("left_button_press", stretch_select_second_left_button_press_cb)
            
def stretch_select_second_left_button_press_cb(doc, event, tool):
    tool.delHandler("mouse_move")
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    _x1, _y1 = tool.getLocation()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    _image = doc.getImage()
    _active_layer = _image.getActiveLayer()
    _dx, _dy = tool.getDistance()
    for _pt in _active_layer.getLayerEntities("point"):
        if _pt.x > _xmax:
            break
        if _pt.inRegion(_xmin, _ymin, _xmax, _ymax):
            _pt.move(_dx, _dy)
    # should do this better . . .
    _init_func = tool.getHandler("initialize")
    tool.reset()
    _init_func(doc, tool)
    _da = doc.getDA()
    _da.flushTempObjects()
    _da.setNeedsDisplay_(True) 

def stretch_horizontal_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the horizontal distance")
    tool.setHandler("initialize", stretch_horizontal_init)
    tool.setHandler("entry_event", stretch_horizontal_entry_event_cb)
    tool.setHandler("left_button_press", stretch_horizontal_first_left_button_press_cb)

def stretch_horizontal_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    tool.setLocation(_x, _y)
    tool.setHandler("left_button_press", stretch_horizontal_second_left_button_press_cb)
    doc.setPrompt("Click a second point to indicate the horizontal distance")

def stretch_horizontal_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    (_x1, _y1) = tool.getLocation()
    tool.setDistance((_x2 - _x1), 0.0)
    tool.setHandler("left_button_press", stretch_select_first_left_button_press_cb)
    tool.delHandler("entry_event")
    doc.setPrompt("Click on or select the objects to stretch")

def stretch_horizontal_entry_event_cb(doc, text, tool):
    if len(text):
        _dist = make_distance(text)
        tool.setDistance(_dist, 0.0)
        tool.setHandler("left_button_press", stretch_select_first_left_button_press_cb)
        tool.delHandler("entry_event")
        doc.setPrompt("Click on or select the points to stretch")

def stretch_vertical_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the vertical distance")
    tool.setHandler("initialize", stretch_vertical_init)
    tool.setHandler("entry_event", stretch_vertical_entry_event_cb)
    tool.setHandler("left_button_press", stretch_vertical_first_left_button_press_cb)

def stretch_vertical_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    tool.setLocation(_x, _y)
    tool.setHandler("left_button_press", stretch_vertical_second_left_button_press_cb)
    doc.setPrompt("Click a second point to indicate the vertical distance")

def stretch_vertical_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    (_x1, _y1) = tool.getLocation()
    tool.setDistance(0.0, (_y2 - _y1))
    tool.setHandler("left_button_press", stretch_select_first_left_button_press_cb)
    tool.delHandler("entry_event")
    doc.setPrompt("Click on or select the points to stretch")

def stretch_vertical_entry_event_cb(doc, text, tool):
    if len(text):
        _dist = make_distance(text)
        tool.setDistance(0.0, _dist)
        tool.setHandler("left_button_press", stretch_select_first_left_button_press_cb)
        tool.delHandler("entry_event")
        doc.setPrompt("Click on or select the points to stretch")


def stretch_free_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the distance to stretch as 'h, v'")
    tool.setHandler("initialize", stretch_free_init)
    tool.setHandler("entry_event", stretch_free_entry_event_cb)
    tool.setHandler("left_button_press", stretch_free_first_left_button_press_cb)

def stretch_free_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    tool.setLocation(_x, _y)
    tool.setHandler("left_button_press", stretch_free_second_left_button_press_cb)
    doc.setPrompt("Click a second point to indicate the vertical distance")

def stretch_free_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    (_x1, _y1) = tool.getLocation()
    tool.setDistance((_x2 - _x1), (_y2 - _y1))
    tool.setHandler("left_button_press", stretch_select_first_left_button_press_cb)
    tool.delHandler("entry_event")
    doc.setPrompt("Click on or select the points to stretch")

def stretch_free_entry_event_cb(doc, text, tool):
    if len(text):
        _str = string.split(text, ",")
        if len(_str) != 2:
            raise ValueError, "Invalid distance (must be h, v): " + text
        _hdist = make_distance(_str[0])
        _vdist = make_distance(_str[1])
        tool.setDistance(_hdist, _vdist)
        tool.setHandler("left_button_press", stretch_select_first_left_button_press_cb)
        tool.delHandler("entry_event")
        doc.setPrompt("Click on or select the points to stretch")


#
# Zoom stuff
#
def zoom_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area to set the zoom location")
    tool.setHandler("initialize", zoom_mode_init)
    tool.setHandler("left_button_press", zoom_first_left_button_press_cb)

def zoom_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    tool.setFirstPoint(_viewLoc.x, _viewLoc.y)
    tool.setLocation(_viewLoc.x, _viewLoc.y)
    _da.setTempPoint(_viewLoc)
    tool.setHandler("left_button_press", zoom_second_left_button_press_cb)
    tool.setHandler("mouse_move", select_mouse_move_cb)
    doc.setPrompt("Move the mouse to select the zoom location")
    
    
def zoom_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x2, _y2) = _da.convertPoint_fromView_(_loc, None)
    (_x1, _y1) = tool.getFirstPoint()
    _x = min(_x1, _x2)
    _y = min(_y1, _y2)
    _w = abs(_x2 - _x1)
    _h = abs(_y2 - _y1)
    _da.setTempObject()
    _da.zoomToRect(((_x, _y), (_w, _h)))
    tool.reset()
    zoom_mode_init(doc, tool)


    


#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# classes for points
#

from __future__ import generators

import math

import entity
import tolerance
import util
import baseobject
import quadtree
import logger

class Point(baseobject.Subpart):
    """A 2-D point Class.

A Point has the following attributes:

x: x-coordinate
y: y-coordinate

A Point object has the following methods:

{get/set}x(): Get/Set the x-coordinate of the Point.
{get/set}y(): Get/Set the y-coordinate of the Point.
{get/set}Coords(): Get/Set both the x and y coordinates of the Point.
move(): Move a Point.
clone(): Return an identical copy of a Point.
inRegion(): Returns True if the point is in some area.
    """
    messages = {
        'moved' : True,
    }
    # member functions

    def __init__(self, x, y=None, **kw):
        """Initialize a Point.

There are two ways to initialize a Point:

Point(xc,yc) - Two arguments, with both arguments being floats
Point((xc,yc)) - A single tuple containing two float objects
        """
        baseobject.Subpart.__init__(self, **kw)
        if isinstance(x, tuple):
            if y is not None:
                raise SyntaxError, "Invalid call to Point()"
            _x, _y = util.tuple_to_two_floats(x)
        elif y is not None:
            _x = x
            if not isinstance(_x, float):
                _x = float(x)
            _y = y
            if not isinstance(_y, float):
                _y = float(y)
        else:
            raise SyntaxError, "Invalid call to Point()."
        self.__x = _x
        self.__y = _y

    def __str__(self):
        return "(%g,%g)" % (self.__x, self.__y)

    def __sub__(self, p):
        """Return the separation between two points.

This function permits the use of '-' to be an easy to read
way to find the distance between two Point objects.
        """
        if not isinstance(p, Point):
            raise TypeError, "Invalid type for Point subtraction: " + str(p)
        _px, _py = p.getCoords()
        return math.hypot((self.__x - _px), (self.__y - _py))

    def __eq__(self, obj):
        """Compare a Point to either another Point or a tuple for equality.
        """
        if not isinstance(obj, (Point,tuple)):
            return False
        if isinstance(obj, Point):
            if obj is self:
                return True
            _x, _y = obj.getCoords()
        else:
            _x, _y = util.tuple_to_two_floats(obj)
        if abs(self.__x - _x) < 1e-10 and abs(self.__y - _y) < 1e-10:
            return True
        return False

    def __ne__(self, obj):
        """Compare a Point to either another Point or a tuple for inequality.
        """
        if not isinstance(obj, (Point,tuple)):
            return True
        if isinstance(obj, Point):
            if obj is self:
                return False
            _x, _y = obj.getCoords()
        else:
            _x, _y = util.tuple_to_two_floats(obj)
        if abs(self.__x - _x) < 1e-10 and abs(self.__y - _y) < 1e-10:
            return False
        return True

    def getValues(self):
        return 'point', entity.Entity.getValues(self), self.__x, self.__y
    
    def getx(self):
        """Return the x-coordinate of a Point.

getx()
        """
        return self.__x

    def setx(self, val):
        """Set the x-coordinate of a Point

setx(val)

The argument "val" must be a float.
        """
        if self.isLocked():
            raise RuntimeError, "Coordinate change not allowed - object locked."
        _v = val
        if not isinstance(_v, float):
            _v = float(val)
        _x = self.__x
        if abs(_x - _v) > 1e-10:
            self.__x = _v
            self.sendMessage('moved', _x, self.__y)
            self.modified()

    x = property(getx, setx, None, "x-coordinate value")

    def gety(self):
        """Return the y-coordinate of a Point.

gety()
        """
        return self.__y

    def sety(self, val):
        """Set the y-coordinate of a Point

sety(val)

The argument "val" must be a float.
        """
        if self.isLocked():
            raise RuntimeError, "Coordinate change not allowed - object locked."
        _v = val
        if not isinstance(_v, float):
            _v = float(val)
        _y = self.__y
        if abs(_y - _v) > 1e-10:
            self.__y = _v
            self.sendMessage('moved', self.__x, _y)
            self.modified()

    y = property(gety, sety, None, "y-coordinate value")

    def getCoords(self):
        """Return the x and y Point coordinates in a tuple.

getCoords()
        """
        return self.__x, self.__y

    def setCoords(self, x, y):
        """Set both the coordinates of a Point.

setCoords(x, y)

Arguments x and y should be float values.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _sx = self.__x
        _sy = self.__y
        if abs(_sx - _x) > 1e-10 or abs(_sy - _y) > 1e-10:
            self.__x = _x
            self.__y = _y
            self.sendMessage('moved', _sx, _sy)
            self.modified()
        
    def move(self, dx, dy):
        """Move a Point.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked():
            raise RuntimeError, "Moving not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x = self.__x
            _y = self.__y
            self.__x = _x + _dx
            self.__y = _y + _dy
            self.sendMessage('moved', _x, _y)
            self.modified()

    def clone(self):
        """Create an identical copy of a Point.

clone()
        """
        return Point(self.__x, self.__y)

    def inRegion(self, xmin, ymin, xmax, ymax, fully=True):
        """Returns True if the Point is within the bounding values.

inRegion(xmin, ymin, xmax, ymax)

The four arguments define the boundary of an area, and the
function returns True if the Point lies within that area.
Otherwise, the function returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        return _xmin < self.__x < _xmax and _ymin < self.__y < _ymax

    def sendsMessage(self, m):
        if m in Point.messages:
            return True
        return baseobject.Subpart.sendsMessage(self, m)

#
# Quadtree Point storage
#

class PointQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def getNodes(self, *args):
        _alen = len(args)
        if _alen != 2:
            raise ValueError, "Expected 2 arguments, got %d" % _alen
        _x, _y = args        
        if not isinstance(_x, float):
            _x = float(args[0])
        if not isinstance(_y, float):
            _y = float(args[1])
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()        
            if ((_x < _xmin) or
                (_y < _ymin) or
                (_x > _xmax) or
                (_y > _ymax)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                if _x < _xmid:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SWNODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NWNODE)
                else:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SENODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NENODE)
                _nodes.append(_node)
            else:
                yield _node

    def addObject(self, obj):
        if not isinstance(obj, Point):
            raise TypeError, "Invalid Point object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCoords()
        _bounds = self.getTreeRoot().getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _x - 1.0
            _ymin = _y - 1.0
            _xmax = _x + 1.0
            _ymax = _y + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _x < _xmin:
                _xmin = _x - 1.0
                _resize = True
            if _x > _xmax:
                _xmax = _x + 1.0
                _resize = True
            if _y < _ymin:
                _ymin = _y - 1.0
                _resize = True
            if _y > _ymax:
                _ymax = _y + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
        for _node in self.getNodes(_x, _y):
            _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, PointQuadtree.movePoint)

    def old_addObject(self, obj):
        if not isinstance(obj, Point):
            raise TypeError, "Invalid Point object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _x - 1.0
            _ymin = _y - 1.0
            _xmax = _x + 1.0
            _ymax = _y + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _x < _xmin:
                _xmin = _x - 1.0
                _resize = True
            if _x > _xmax:
                _xmax = _x + 1.0
                _resize = True
            if _y < _ymin:
                _ymin = _y - 1.0
                _resize = True
            if _y > _ymax:
                _ymax = _y + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        while True:
            if ((_x > _xmax) or
                (_y > _ymax) or
                (_x < _xmin) or
                (_y < _ymin)):
                print "invalid node:"
                print "boundary: (%g, %g) -> (%g, %g)" % (_xmin, _ymin,
                                                          _xmax, _ymax)
                print "point: (%g, %g)" % (_x, _y)
                raise ValueError, "Failed to add point: " + `obj`
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                if _x < _xmid:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SWNODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NWNODE)
                else:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SENODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NENODE)
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            else:
                _node.addObject(obj)
                obj.connect('moved', self, PointQuadtree.movePoint)
                break
        quadtree.Quadtree.addObject(self, obj)

    def delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCoords()
        for _node in self.getNodes(_x, _y):
            _node.delObject(obj)
            _parent = _node.getParent()
            if _parent is not None:
                self.purgeSubnodes(_parent)
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)

    def old_delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCoords()
        _node = self.getTreeRoot()
        while True:
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x > _xmax) or
                (_y > _ymax) or
                (_x < _xmin) or
                (_y < _ymin)):
                print "invalid node:"
                print "boundary: (%g, %g) -> (%g, %g)" % (_xmin, _ymin,
                                                          _xmax, _ymax)
                print "point: (%g, %g)" % (_x, _y)
                raise ValueError, "Failed to delete point: " + `obj`
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                if _x < _xmid:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SWNODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NWNODE)
                else:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SENODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NENODE)
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            else:
                _node.delObject(obj)
                _parent = _node.getParent()
                if _parent is not None:
                    self.purgeSubnodes(_parent)
                break
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)

    def find(self, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _t = tolerance.TOL
        if _alen > 2 :
            _t = tolerance.toltest(args[2])
        if not len(self):
            return None
        _nodes = [self.getTreeRoot()]
        _p = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_y < (_ymin - _t)) or
                (_x > (_xmax + _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _xmin, _ymin, _xmax, _ymax = _subnode.getBoundary()
                    if ((_xmin - _t) < _x < (_xmax + _t) and
                        (_ymin - _t) < _y < (_ymax + _t)):
                        _nodes.append(_subnode)
            else:
                for _pt in _node.getObjects():
                    _xp, _yp = _pt.getCoords()
                    _xsep = _xp - _x
                    _ysep = _yp - _y
                    if abs(_xsep) < _t and abs(_ysep) < _t:
                        if _p is None:
                            _p = _pt
                        else:
                            if math.hypot(_xsep, _ysep) < (_p - _pt):
                                _p = _pt
        return _p

    def movePoint(self, obj, *args):
        if obj not in self:
            raise ValueError, "Point not stored in Quadtree: " + `obj`
        _alen = len(args)
        if len(args) < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        for _node in self.getNodes(_x, _y):
            _node.delObject(obj)
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def old_movePoint(self, obj, *args):
        if obj not in self:
            raise ValueError, "Point not stored in Quadtree: " + `obj`
        _alen = len(args)
        if len(args) < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _node = self.getTreeRoot()
        _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
        while True:
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                if _x < _xmid:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SWNODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NWNODE)
                else:
                    if _y < _ymid:
                        _node = _node.getSubnode(quadtree.QTreeNode.SENODE)
                    else:
                        _node = _node.getSubnode(quadtree.QTreeNode.NENODE)
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            else:
                _node.delObject(obj)
                quadtree.Quadtree.delObject(self, obj)
                obj.disconnect(self)
                break
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        return self.find(x, y, tol)

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _pts = []        
        if not len(self):
            return _pts
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _pt in _node.getObjects():
                    if _pt.inRegion(_xmin, _ymin, _xmax, _ymax):
                        _pts.append(_pt)
        return _pts

#
# Point history class
#

class PointLog(logger.Logger):
    def __init__(self, p):
        if not isinstance(p, Point):
            raise TypeError, "Invalid point: " + `p`
        logger.Logger.__init__(self)
        self.__p = p
        p.connect('moved', self, PointLog.movePoint)

    def detatch(self):
        self.__p.disconnect(self)
        self.__p = None
        
    def movePoint(self, p, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        self.saveUndoData('moved', _x, _y)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if _alen == 0:
            raise ValueError, "No arguments to execute()"
        _p = self.__p
        _op = args[0]
        if _op == 'moved':        
            if _alen < 3:
                raise ValueError, "Invalid argument count: %d" % _alen
            _x = args[1]
            if not isinstance(_x, float):
                _x = float(args[1])
            _y = args[2]
            if not isinstance(_y, float):
                _y = float(args[2])
            _px = _p.x
            _py = _p.y
            if undo:
                self.saveRedoData('moved', _px, _py)
            else:
                self.saveUndoData('moved', _px, _py)
            self.ignore('moved')
            try:
                if undo:
                    _p.startUndo(True)
                    try:
                        _p.setCoords(_x, _y)
                    finally:
                        _p.endUndo()
                    if abs(_x - _px) > 1e-10 or abs(_y - _py) > 1e-10:
                        _p.sendMessage('moved', _px, _py)
                else:
                    _p.setCoords(_x, _y)
            finally:
                self.receive('moved')
        else:
            raise ValueError, "Unexpected operation: %s" % _op
        



#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# a construction line defined by two points
#

from __future__ import generators

import math

import conobject
import entity
import tolerance
import point
import quadtree
import logger

class CLine(conobject.ConstructionObject):
    """A class for construction lines defined by two distinct Points.

A CLine object is derived from the conobject class, so it shares
the functionality of that class. In addition, a CLine instance
has two attributes:

p1: A Point object representing the first keypoint
p2: A Point object representing the second keypoint

A CLine has the following methods:

getKeypoints(): Return the two points the CLine is defined by.
{get/set}P1: Get/Set the first keypoint of the CLine.
{get/set}P2: Get/Set the second keypoint of the CLine.
move(): Move a CLine
mapCoords(): Return the nearest point on a CLine to a coordinate pair.
inRegion(): Return whether the CLine passes through a bounded region.
clone(): Return an identical copy of a CLine.
    """
    messages = {
        'moved' : True,
        'keypoint_changed' : True
    }
    def __init__(self, p1, p2, **kw):
        """Initialize a CLine object.

CLine(p1, p2)

Both arguments are Point objects that the CLine passes through.
        """
        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)
        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 is _p2:
            raise ValueError, "A CLine must have two different keypoints."
        conobject.ConstructionObject.__init__(self, **kw)
        self.__p1 = _p1
        _p1.storeUser(self)
        _p1.connect('moved', self, CLine.movePoint)
        self.__p2 = _p2
        _p2.storeUser(self)
        _p2.connect('moved', self, CLine.movePoint)

    def __eq__(self, obj):
        """Compare one CLine to another for equality.
        """
        if not isinstance(obj, CLine):
            return False
        if obj is self:
            return True
        _sp1, _sp2 = self.getKeypoints()
        _op1, _op2 = obj.getKeypoints()
        _sv = abs(_sp1.x - _sp2.x) < 1e-10
        _sh = abs(_sp1.y - _sp2.y) < 1e-10
        _ov = abs(_op1.x - _op2.x) < 1e-10
        _oh = abs(_op1.y - _op2.y) < 1e-10
        _val = False
        if _sv and _ov: # both vertical
            if abs(_sp1.x - _op1.x) < 1e-10:
                _val = True
        elif _sh and _oh: # both horizontal
            if abs(_sp1.y - _op1.y) < 1e-10:
                _val = True
        else:
            if (not (_sv or _sh)) and (not (_ov or _oh)):
                _sx1, _sy1 = _sp1.getCoords()
                _sx2, _sy2 = _sp2.getCoords()
                _ox1, _oy1 = _op1.getCoords()
                _ox2, _oy2 = _op2.getCoords()
                _ms = (_sy2 - _sy1)/(_sx2 - _sx1)
                _bs = _sy1 - (_ms * _sx1)
                _ty = (_ms * _ox1) + _bs
                if abs(_ty - _oy1) < 1e-10:
                    _ty = (_ms * _ox2) + _bs
                    if abs(_ty - _oy2) < 1e-10:
                        _val = True
        return _val

    def __ne__(self, obj):
        """Compare one CLine to another for inequality.
        """
        if not isinstance(obj, CLine):
            return True
        if obj is self:
            return False
        _sp1, _sp2 = self.getKeypoints()
        _op1, _op2 = obj.getKeypoints()
        _sv = abs(_sp1.x - _sp2.x) < 1e-10
        _sh = abs(_sp1.y - _sp2.y) < 1e-10
        _ov = abs(_op1.x - _op2.x) < 1e-10
        _oh = abs(_op1.y - _op2.y) < 1e-10
        _val = True
        if _sv and _ov: # both vertical
            if abs(_sp1.x - _op1.x) < 1e-10:
                _val = False
        elif _sh and _oh: # both horizontal
            if abs(_sp1.y - _op1.y) < 1e-10:
                _val = False
        else:
            if (not (_sv or _sh)) and (not (_ov or _oh)):
                _sx1, _sy1 = _sp1.getCoords()
                _sx2, _sy2 = _sp2.getCoords()
                _ox1, _oy1 = _op1.getCoords()
                _ox2, _oy2 = _op2.getCoords()
                _ms = (_sy2 - _sy1)/(_sx2 - _sx1)
                _bs = _sy1 - (_ms * _sx1)
                _ty = (_ms * _ox1) + _bs
                if abs(_ty - _oy1) < 1e-10:
                    _ty = (_ms * _ox2) + _bs
                    if abs(_ty - _oy2) < 1e-10:
                        _val = False
        return _val

    def __str__(self):
        return "Construction Line through %s and %s" % (self.__p1, self.__p2)

    def getValues(self):
        _p1id = self.__p1.getID()
        _p2id = self.__p2.getID()
        return 'cline', entity.Entity.getValues(self), _p1id, _p2id

    def getKeypoints(self):
        """Return the two keypoints of this CLine.

getKeypoints()
        """
        return self.__p1, self.__p2

    def getP1(self):
        """Return the first keypoint Point of the CLine.

getP1()
        """
        return self.__p1

    def setP1(self, p):
        """Set the first keypoint Point of the CLine.

setP1(p)

Argument p must be a Point.
        """
        if self.isLocked():
            raise RuntimeError, "Setting keypoint not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid keypoint: " + str(p)
        if p is self.__p2 or p == self.__p2:
            raise ValueError, "CLines must have two different keypoints."
        _kp = self.__p1
        if _kp is not p:
            _kp.disconnect(self)
            _kp.freeUser(self)
            self.__p1 = p
            self.sendMessage('keypoint_changed', _kp, p)
            p.storeUser(self)
            p.connect('moved', self, CLine.movePoint)
            if abs(_kp.x - p.x) > 1e-10 or abs(_kp.y - p.y) > 1e-10:
                _x, _y = self.__p2.getCoords()
                self.sendMessage('moved', _kp.x, _kp.y, _x, _y)
            self.modified()

    p1 = property(getP1, setP1, None, "First keypoint of the CLine.")

    def getP2(self):
        """Return the second keypoint Point of the CLine.

getP2()
        """
        return self.__p2

    def setP2(self, p):
        """Set the second keypoint Point of the CLine.

setP2(p)

Argument p must be a Point.
        """
        if self.isLocked():
            raise RuntimeError, "Setting keypoint not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid keypoint: " + `p`
        if p is self.__p1 or p == self.__p1:
            raise ValueError, "CLines must have two different keypoints."
        _kp = self.__p2
        if _kp is not p:
            _kp.disconnect(self)
            _kp.freeUser(self)
            self.__p2 = p
            self.sendMessage('keypoint_changed', _kp, p)
            p.storeUser(self)
            p.connect('moved', self, CLine.movePoint)
            if abs(_kp.x - p.x) > 1e-10 or abs(_kp.y - p.y) > 1e-10:
                _x, _y = self.__p1.getCoords()
                self.sendMessage('moved', _x, _y, _kp.x, _kp.y)
            self.modified()

    p2 = property(getP2, setP2, None, "Second keypoint of the CLine.")

    def move(self, dx, dy):
        """Move a CLine.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked() or self.__p1.isLocked() or self.__p2.isLocked():
            raise RuntimeError, "Moving CLine not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x1, _y1 = self.__p1.getCoords()
            _x2, _y2 = self.__p2.getCoords()
            self.ignore('moved')
            self.__p1.move(_dx, _dy)
            self.__p2.move(_dx, _dy)
            self.receive('moved')
            self.sendMessage('moved', _x1, _y1, _x2, _y2)
            self.modified()

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the CLine to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to a
actual Point on the CLine. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _sqlen = pow((_x2 - _x1), 2) + pow((_y2 - _y1), 2)
        if _sqlen < 1e-10: # both points the same
            raise RuntimeError, "CLine points coincident."
        _r = ((_x - _x1)*(_x2 - _x1) + (_y - _y1)*(_y2 - _y1))/_sqlen
        _px = _x1 + _r * (_x2 - _x1)
        _py = _y1 + _r * (_y2 - _y1)
        if abs(_px - _x) < _t and abs(_py - _y) < _t:
           return _px, _py
        return None

    def getProjection(self, x, y):
        """Find the projection point of some coordinates on the CLine.

getProjection(x, y)

Arguments "x" and "y" should be float values.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _sqlen = pow((_x2 - _x1), 2) + pow((_y2 - _y1), 2)
        _rn = ((_x - _x1) * (_x2 - _x1)) + ((_y - _y1) * (_y2 - _y1))
        _r = _rn/_sqlen
        _px = _x1 + _r * (_x2 - _x1)
        _py = _y1 + _r * (_y2 - _y1)
        return _px, _py

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a CLine passes through a region.

isRegion(xmin, ymin, xmax, ymax)

The four arguments define the boundary of an area, and the
function returns True if the CLine passes within the area.
Otherwise, the function returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + str(fully)
        if fully:
            return False
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _xdiff = _x2 - _x1
        _ydiff = _y2 - _y1
        _val = False
        if _xmin < _x1 < _xmax and _ymin < _y1 < _ymax:
            _val = True
        elif _xmin < _x2 < _xmax and _ymin < _y2 < _ymax:
            _val = True
        elif abs(_xdiff) < 1e-10: # vertical line
            if _xmin < _x1 < _xmax:
                _val = True
        elif abs(_ydiff) < 1e-10: # horizontal line
            if _ymin < _y1 < _ymax:
                _val = True
        else:
            _slope = _ydiff/_xdiff
            _yint = _y1 - _slope*_x1
            if _ymin < (_slope*_xmin + _yint) < _ymax: # hits left side
                _val = True
            elif _ymin < (_slope*_xmax + _yint) < _ymax: # hits right side
                _val = True
            else: # hits bottom - no need to check top ...
                _xymin = (_ymin - _yint)/_slope
                if _xmin < _xymin < _xmax:
                    _val = True
        return _val

    def movePoint(self, p, *args):
        _plen = len(args)
        if _plen < 2:
            raise ValueError, "Invalid argument count: %d" % _plen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _p1 = self.__p1
        _p2 = self.__p2
        if p is _p1:
            _x1 = _x
            _y1 = _y
            _x2, _y2 = _p2.getCoords()
            if abs(_p1.x - _x2) < 1e-10 and abs(_p1.y - _y2) < 1e-10:
                raise RuntimeError, "CLine points coincident."
        elif p is _p2:
            _x1, y1 = _p1.getCoords()
            _x2 = _x
            _y2 = _y
            if abs(_p2.x - _x1) < 1e-10 and abs(_p2.y - _y1) < 1e-10:
                raise RuntimeError, "CLine points coincident."
        else:
            raise ValueError, "Unexpected CLine keypoint: " + `p`
        self.sendMessage('moved', _x1, _y1, _x2, _y2)
        self.modified()

    def clone(self):
        """Create an identical copy of a CLine.

clone()
        """
        _cp1 = self.__p1.clone()
        _cp2 = self.__p2.clone()
        return CLine(_cp1, _cp2)

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _p1, _p2 = self.getKeypoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _coords = None
        if abs(_x2 - _x1) < 1e-10: # vertical
            if _xmin < _x1 < _xmax:
                _coords = (_x1, _ymin, _x1, _ymax)
        elif abs(_y2 - _y1) < 1e-10: # horiztonal
            if _ymin < _y1 < _ymax:
                _coords = (_xmin, _y1, _xmax, _y1)
        else:
            #
            # the CLine can be parameterized as
            #
            # x = u * (x2 - x1) + x1
            # y = u * (y2 - y1) + y1
            #
            # for u = 0, x => x1, y => y1
            # for u = 1, x => x2, y => y2
            #
            # The following is the Liang-Barsky Algorithm
            # for segment clipping modified slightly for
            # construction lines
            #
            _dx = _x2 - _x1
            _dy = _y2 - _y1
            # print "dx: %g; dy: %g" % (_dx, _dy)
            _P = [-_dx, _dx, -_dy, _dy]
            _q = [(_x1 - _xmin), (_xmax - _x1), (_y1 - _ymin), (_ymax - _y1)]
            _u1 = None
            _u2 = None
            _valid = True
            for _i in range(4):
                # print "i: %d" % _i
                _pi = _P[_i]
                _qi = _q[_i]
                # print "p[i]: %g; q[i]: %g" % (_pi, _qi)
                if abs(_pi) < 1e-10: # this should be caught earlier ...
                    if _qi < 0.0:
                        _valid = False
                        break
                else:
                    _r = _qi/_pi
                    # print "r: %g" % _r
                    if _pi < 0.0:
                        # print "testing u1 ..."
                        if _u2 is not None and _r > _u2:
                            # print "r > u2 (%g)" % _u2
                            _valid = False
                            break
                        if _u1 is None or _r > _u1:
                            # print "setting u1 = r"
                            _u1 = _r
                    else:
                        # print "testing u2 ..."
                        if _u1 is not None and _r < _u1:
                            # print "r < u1 (%g)" % _u1
                            _valid = False
                            break
                        if _u2 is None or _r < _u2:
                            # print "setting u2 = r"
                            _u2 = _r
            if _valid:
                _coords = (((_u1 * _dx) + _x1),
                           ((_u1 * _dy) + _y1),
                           ((_u2 * _dx) + _x1),
                           ((_u2 * _dy) + _y1))
        return _coords

    def sendsMessage(self, m):
        if m in CLine.messages:
            return True
        return conobject.ConstructionObject.sendsMessage(self, m)

def intersect_region(cl, xmin, ymin, xmax, ymax):
    if not isinstance(cl, CLine):
        raise TypeError, "Invalid CLine: " + str(cl)
    _xmin = xmin
    if not isinstance(_xmin, float):
        _xmin = float(xmin)
    _ymin = ymin
    if not isinstance(_ymin, float):
        _ymin = float(ymin)
    _xmax = xmax
    if not isinstance(_xmax, float):
        _xmax = float(xmax)
    if _xmax < _xmin:
        raise ValueError, "Illegal values: xmax < xmin"
    _ymax = ymax
    if not isinstance(_ymax, float):
        _ymax = float(ymax)
    if _ymax < _ymin:
        raise ValueError, "Illegal values: ymax < ymin"
    _p1, _p2 = cl.getKeypoints()
    _p1x, _p1y = _p1.getCoords()
    _p2x, _p2y = _p2.getCoords()
    _x1 = _y1 = _x2 = _y2 = None
    if abs(_p2x - _p1x) < 1e-10: # vertical
        if _xmin < _p1x < _xmax:
            _x1 = _p1x
            _y1 = _ymin
            _x2 = _p1x
            _y2 = _ymax
    elif abs(_p2y - _p1y) < 1e-10: # horiztonal
        if _ymin < _p1y < _ymax:
            _x1 = _xmin
            _y1 = _p1y
            _x2 = _xmax
            _y2 = _p1y
    else:
        _slope = (_p2y - _p1y)/(_p2x - _p1x)
        _yint = _p1y - (_p1x * _slope)
        #
        # find y for x = xmin
        #
        _yt = (_slope * _xmin) + _yint
        if _ymin < _yt < _ymax:
            # print "hit at y for x=xmin"
            _x1 = _xmin
            _y1 = _yt
        #
        # find y for x = xmax
        #
        _yt = (_slope * _xmax) + _yint
        if _ymin < _yt < _ymax:
            # print "hit at y for x=xmax"
            if _x1 is None:
                _x1 = _xmax
                _y1 = _yt
            else:
                _x2 = _xmax
                _y2 = _yt
        if _x2 is None:
            #
            # find x for y = ymin
            #
            _xt = (_ymin - _yint)/_slope
            if _xmin < _xt < _xmax:
                # print "hit at x for y=ymin"
                if _x1 is None:
                    _x1 = _xt
                    _y1 = _ymin
                else:
                    _x2 = _xt
                    _y2 = _ymin
        if _x2 is None:
            #
            # find x for y = ymax
            #
            _xt = (_ymax - _yint)/_slope
            if _xmin < _xt < _xmax:
                # print "hit at x for y=ymax"
                if _x1 is None:
                    _x1 = _xt
                    _y1 = _ymax
                else:
                    _x2 = _xt
                    _y2 = _ymax
    return _x1, _y1, _x2, _y2

#
# Quadtree CLine storage
#

class CLineQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def getNodes(self, *args):
        _alen = len(args)
        if _alen != 4:
            raise ValueError, "Expected 4 arguments, got %d" % _alen
        _x1, _y1, _x2, _y2 = args
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _h = abs(_y2 - _y1) < 1e-10
        _v = abs(_x2 - _x1) < 1e-10
        if _h and _v: # both coords are identical
            raise ValueError, "CLine singularity - identical coords."
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = False
                if _v:
                    if _x1 < _xmin or _x1 > _xmax:
                        continue
                    if _x1 < _xmid: # cline on left
                        _sw = _nw = True
                    else:
                        _se = _ne = True
                elif _h:
                    if _y1 < _ymin or _y1 > _ymax:
                        continue
                    if _y1 < _ymid: # cline below
                        _sw = _se = True
                    else:
                        _nw = _ne = True
                else:
                    _ne = _nw = _sw = _se = True
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                yield _node

    def addObject(self, obj):
        if not isinstance(obj, CLine):
            raise TypeError, "Invalid CLine object: " + `obj`
        if obj in self:
            return
        _p1, _p2 = obj.getKeypoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _bounds = self.getTreeRoot().getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _sxmin - 1.0
            _ymin = _symin - 1.0
            _xmax = _sxmax + 1.0
            _ymax = _symax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _sxmin < _xmin:
                _xmin = _sxmin - 1.0
                _resize = True
            if _sxmax > _xmax:
                _xmax = _sxmax + 1.0
                _resize = True
            if _symin < _ymin:
                _ymin = _symin - 1.0
                _resize = True
            if _symax > _ymax:
                _ymax = _symax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
        for _node in self.getNodes(_x1, _y1, _x2, _y2):
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, CLineQuadtree.moveCLine)

    def old_addObject(self, obj):
        if not isinstance(obj, CLine):
            raise TypeError, "Invalid CLine object: " + `obj`
        if obj in self:
            return
        _p1, _p2 = obj.getKeypoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _horiz = abs(_y1 - _y2) < 1e-10
        _vert = abs(_x1 - _x2) < 1e-10
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _sxmin - 1.0
            _ymin = _symin - 1.0
            _xmax = _sxmax + 1.0
            _ymax = _symax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _sxmin < _xmin:
                _xmin = _sxmin - 1.0
                _resize = True
            if _sxmax > _xmax:
                _xmax = _sxmax + 1.0
                _resize = True
            if _symin < _ymin:
                _ymin = _symin - 1.0
                _resize = True
            if _symax > _ymax:
                _ymax = _symax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _vert:
                    if _sxmax < _xmid: # cline on left
                        _ne = _se = False
                    else:
                        _nw = _sw = False
                if _horiz:
                    if _symax < _ymid: # cline below
                        _nw = _ne = False
                    else:
                        _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, CLineQuadtree.moveCLine)

    def delObject(self, obj):
        if obj not in self:
            return
        _p1, _p2 = obj.getKeypoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _pdict = {}
        for _node in self.getNodes(_x1, _y1, _x2, _y2):
            _node.delObject(obj) # cline may not be in the node ...
            _parent = _node.getParent()
            if _parent is not None:
                _pid = id(_parent)
                if _pid not in _pdict:
                    _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def old_delObject(self, obj):
        if obj not in self:
            return
        _p1, _p2 = obj.getKeypoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _xmin = _ymin = _xmax = _ymax = None
        # _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        # _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _horiz = abs(_y1 - _y2) < 1e-10
        _vert = abs(_x1 - _x2) < 1e-10
        _pdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _vert:
                    if _sxmax < _xmid: # cline on left side
                        _ne = _se = False
                    else:
                        _nw = _sw = False
                if _horiz:
                    if _symax < _ymid: # cline below
                        _nw = _ne = False
                    else:
                        _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # cline may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        _t = tolerance.TOL
        if _alen > 4:
            _t = tolerance.toltest(args[4])
        if not len(self):
            return None
        # _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        # _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _horiz = abs(_y1 - _y2) < 1e-10
        _vert = abs(_x1 - _x2) < 1e-10
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _cdict = {}
        _bailout = False
        _cline = None
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _vert:
                    if _sxmax < (_xmid - _t): # cline on left side
                        _ne = _se = False
                    if _sxmax > (_xmid + _t): # cline on right side
                        _nw = _sw = False
                if _horiz:
                    if _symax < (_ymid - _t): # cline below
                        _nw = _ne = False
                    if _symax > (_ymid + _t): # cline above
                        _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                for _c in _node.getObjects():
                    _cid = id(_c)
                    if _cid not in _cdict:
                        _cx, _cy = _c.getProjection(_x1, _y1)
                        if math.hypot((_cx - _x1), (_cy - _y1)) < _t:
                            _cx, _cy = _c.getProjection(_x2, _y2)
                            if math.hypot((_cx - _x2), (_cy - _y2)) < _t:
                                _bailout = True
                                _cline = _c
                                break
                        _cdict[_cid] = True
            if _bailout:
                break
        return _cline

    def moveCLine(self, obj, *args):
        if obj not in self:
            raise ValueError, "CLine not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        for _node in self.getNodes(_x1, _y1, _x2, _y2):
            _node.delObject(obj) # cline may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def old_moveCLine(self, obj, *args):
        if obj not in self:
            raise ValueError, "CLine not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        # _sxmin = min(_x1, _x2)
        _sxmax = max(_x1, _x2)
        # _symin = min(_y1, _y2)
        _symax = max(_y1, _y2)
        _horiz = abs(_y1 - _y2) < 1e-10
        _vert = abs(_x1 - _x2) < 1e-10
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if _node.hasSubnodes():
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _vert:
                    if _sxmax < _xmid: # cline on left side
                        _ne = _se = False
                    else:
                        _nw = _sw = False
                if _horiz:
                    if _symax < _ymid: # cline below
                        _nw = _ne = False
                    else:
                        _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # cline may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _cline = _tsep = None
        _cdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _c in _node.getObjects():
                    _cid = id(_c)
                    if _cid not in _cdict:
                        _cx, _cy = _c.getProjection(_x, _y)
                        if abs(_cx - _x) < _t and abs(_cy - _y) < _t:
                            _sep = math.hypot((_cx - _x), (_cy - _y))
                            if _tsep is None:
                                _tsep = _sep
                                _cline = _c
                            else:
                                if _sep < _tsep:
                                    _tsep = _sep
                                    _cline = _c
        return _cline

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _clines = []
        if not len(self):
            return _clines
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _cdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _nodes.append(_subnode)
            else:
                for _cline in _node.getObjects():
                    _cid = id(_cline)
                    if _cid not in _cdict:
                        if _cline.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _clines.append(_cline)
                        _cdict[_cid] = True
        return _clines

#
# CLine history class
#

class CLineLog(logger.Logger):
    def __init__(self, c):
        if not isinstance(c, CLine):
            raise TypeError, "Invalid CLine: " + `c`
        logger.Logger.__init__(self)
        self.__c = c
        c.connect('keypoint_changed', self, CLineLog.keypointChange)

    def detatch(self):
        self.__c.disconnect(self)
        self.__c = None

    def keypointChange(self, c, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _old = args[0]
        if not isinstance(_old, point.Point):
            raise TypeError, "Invalid old keypoint: " + `_old`
        _oid = _old.getID()
        _new = args[1]
        if not isinstance(_new, point.Point):
            raise TypeError, "Invalid new keypoint: " + `_new`
        _nid = _new.getID()
        self.saveUndoData('keypoint_changed', _oid, _nid)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if len(args) == 0:
            raise ValueError, "No arguments to execute()"
        _c = self.__c
        _p1, _p2 = _c.getKeypoints()
        _op = args[0]
        if _op == 'keypoint_changed':
            if _alen < 3:
                raise ValueError, "Invalid argument count: %d" % _alen
            _oid = args[1]
            _nid = args[2]
            if undo:
                self.saveRedoData('keypoint_changed', _oid, _nid)
                self.ignore('keypoint_changed')
                try:
                    _parent = _c.getParent()
                    if _parent is None:
                        raise ValueError, "CLine has no parent - cannot undo"
                    _op = _parent.getObject(_oid)
                    if _op is None or not isinstance(_op, point.Point):
                        raise ValueError, "Old keypoint missing: id=%d" % _oid
                    _x1, _y1 = _p1.getCoords()
                    _x2, _y2 = _p2.getCoords()
                    _moved = False
                    _c.startUndo(True)
                    try:
                        if _p1.getID() == _nid:
                            _c.setP1(_op)
                            if (abs(_op.x - _x1) > 1e-10 or
                                abs(_op.y - _y1) > 1e-10):
                                _moved = True
                        elif _p2.getID() == _nid:
                            _c.setP2(_op)
                            if (abs(_op.x - _x2) > 1e-10 or
                                abs(_op.y - _y2) > 1e-10):
                                _moved = True
                        else:
                            raise ValueError, "Unexpected keypoint ID: %d" % _nid
                    finally:
                        _c.endUndo()
                    if _moved:
                        _c.sendMessage('moved', _x1, _y1, _x2, _y2)
                finally:
                    self.receive('keypoint_changed')
            else:
                self.saveUndoData('keypoint_changed', _oid, _nid)
                _parent = _c.getParent()
                if _parent is None:
                    raise ValueError, "CLine has no parent - cannot undo"
                _np = _parent.getObject(_nid)
                if _np is None or not isinstance(_np, point.Point):
                    raise ValueError, "New keypoint missing: id=%d" % _nid
                self.ignore('keypoint_changed')
                try:
                    if _p1.getID() == _oid:
                        _c.setP1(_np)
                    elif _p2.getID() == _oid:
                        _c.setP2(_np)
                    else:
                        raise ValueError, "Unexpected keypoint ID: %d" % _oid
                finally:
                    self.receive('keypoint_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op



#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import struct
import sys
_debug = 0
dbg_handle = sys.stdout

#
# bitstream reading functions for DWG files
#
# these functions are used in R13/R14/R15 file decoding
#
# data: an array.array instance of unsigned bytes ("B")
# offset: the current bit offset where the value begins
#

def read_extended_data(data, offset):
    _bitpos = offset
    _extdata = []
    while True:
        _bitpos, _size = get_bit_short(data, _bitpos)
        if _size == 0:
            break
        _bitpos, _handle = get_handle(data, _bitpos)
        _eedata = []
        while (_size > 0):
            _bitpos, _cb = get_raw_char(data, _bitpos)
            _size = _size - 1
            if _cb == 0x0: # string
                _bitpos, _len = get_raw_char(data, _bitpos)
                _bitpos, _cp = get_raw_short(data, _bitpos)
                _chars = []
                for _i in range(_len):
                    _bitpos, _char = get_raw_char(data, _bitpos)
                    _chars.append(chr(_char))
                _eedata.append("".join(_chars))
                _size = _size - _len - 3
            elif _cb == 0x1:
                raise ValueError, "invalid EEX code byte: 0x1"
            elif _cb == 0x2: # either '{' or '}'
                _bitpos, _char = get_raw_char(data, _bitpos)
                if _char == 0x0:
                    _eedata.append("{")
                elif _char == 0x1:
                    _eedata.append("}")
                else:
                    raise ValueError, "Unexpected EEX char: %#02x" % _char
                _size = _size - 1
            elif (_cb == 0x3 or # layer table reference
                  _cb == 0x5): # entity handle reference
                _chars = []
                for _i in range(8):
                    _bitpos, _char = get_raw_char(data, _bitpos)
                    _chars.append(_char)
                _eedata.append(tuple(_chars)) # this seems odd ...
                _size = _size - 8
            elif _cb == 0x4: # binary data
                _bitpos, _len = get_raw_char(data, _bitpos)
                _chars = []
                for _i in range(_len):
                    _bitpos, _char = get_raw_char(data, _bitpos)
                    _chars.append(_char)
                _eedata.append(_chars)
                _size = _size - _len - 1
            elif (0xa <= _cb <= 0xd): # three doubles
                _bitpos, _d1 = get_raw_double(data, _bitpos)
                _bitpos, _d2 = get_raw_double(data, _bitpos)
                _bitpos, _d3 = get_raw_double(data, _bitpos)
                _eedata.append((_d1, _d2, _d3))
                _size = _size - 24
            elif (0x28 <= _cb <= 0x2a): # one double
                _bitpos, _d = get_raw_double(data, _bitpos)
                _eedata.append(_d)
                _size = _size - 8
            elif _cb == 0x46: # short int
                _bitpos, _short = get_raw_short(data, _bitpos)
                _eedata.append(_short)
                _size = _size - 2
            elif _cb == 0x47: # long int
                _bitpos, _long = get_raw_long(data, _bitpos)
                _eedata.append(_long)
                _size = _size - 4
            else:
                raise ValueError, "Unexpected code byte: %#02x" % _cb
        _extdata.append((_handle, _eedata))
    return _bitpos, _extdata

def get_default_double(data, offset, defval):
    _flags = get_bits(data, 2, offset)
    _read = 2
    if (_flags == 0x0):
        _val = defval
    else:
        _offset = offset + 2
        if (_flags == 0x3):
            _dstr = get_bits(data, 64, _offset)
            _val = struct.unpack('<d', _dstr)[0]
            _read = 66
        else:
            _dstr = list(struct.pack('<d', defval))
            if (_flags == 0x1):
                _dd = get_bits(data, 32, _offset)
                _dstr[0] = _dd[0]
                _dstr[1] = _dd[1]
                _dstr[2] = _dd[2]
                _dstr[3] = _dd[3]
                _read = 34
            else: # flags == 0x2
                _dd = get_bits(data, 48, _offset)
                _dstr[4] = _dd[0]
                _dstr[5] = _dd[1]
                _dstr[0] = _dd[2]
                _dstr[1] = _dd[3]
                _dstr[2] = _dd[4]
                _dstr[3] = _dd[5]
                _read = 50
            _val = struct.unpack('<d', "".join(_dstr))[0]
    return (offset + _read), _val

def get_bit_double(data, offset):
    _type = get_bits(data, 2, offset)
    _read = 2
    _val = None
    if _type == 0x00:
        _db = get_bits(data, 64, (offset + 2))
        _val = struct.unpack('<d', _db)[0]
        _read = 66
    elif _type == 0x01:
        _val = 1.0
    elif _type == 0x02:
        _val = 0.0
    else:
        raise ValueError, "bad type at bit offset %d: 0x3" % offset
    return (offset + _read), _val

def get_raw_double(data, offset):
    _db = get_bits(data, 64, offset)
    _val = struct.unpack('<d', _db)[0]
    return (offset + 64), _val

def get_bit_short(data, offset):
    _type = get_bits(data, 2, offset)
    _read = 2
    _val = None
    if _type == 0x00:
        _sh = get_bits(data, 16, (offset + 2))
        _val = struct.unpack('<h', _sh)[0]
        _read = 18
    elif _type == 0x01:
        _val = get_bits(data, 8, (offset + 2))
        _read = 10
    elif _type == 0x02:
        _val = 0
    elif _type == 0x03:
        _val = 256
    return (offset + _read), _val

def get_raw_short(data, offset):
    _sh = get_bits(data, 16, offset)
    _val = struct.unpack('<h', _sh)[0]
    return (offset + 16), _val

def get_bit_long(data, offset):
    _type = get_bits(data, 2, offset)
    _read = 2
    _val = None
    if _type == 0x0:
        _long = get_bits(data, 32, (offset + 2))
        _val = struct.unpack('<l', _long)[0]
        _read = 34
    elif _type == 0x01:
        _val = get_bits(data, 8, (offset + 2))
        _read = 10
    elif _type == 0x02:
        _val = 0
    else:
        raise ValueError, "Bad type at bit offset %d: 0x03" % offset
    return (offset + _read), _val

def get_raw_long(data, offset):
    _long = get_bits(data, 32, offset)
    _val = struct.unpack('<l', _long)[0]
    return (offset + 32), _val

def get_raw_char(data, offset):
    _char = get_bits(data, 8, offset)
    return (offset + 8), _char

def get_modular_char(data, offset):
    _bytes = []
    _read = True
    _offset = offset
    _fac = 1
    while _read:
        _byte = get_bits(data, 8, _offset)
        _offset = _offset + 8
        if not (_byte & 0x80): # final byte has 0 in high bit
            _read = False
            if (_byte & 0x40): # negation
                _fac = -1
                _byte = _byte & 0xbf # turn off negative bit
        _bytes.append(_byte & 0x7f) # turn off high bit
    if len(_bytes) == 1:
        _val = _bytes[0]
    elif len(_bytes) == 2:
        _val = _bytes[0] | (_bytes[1] << 7)
    elif len(_bytes) == 3: # possible?
        _val = _bytes[0] | (_bytes[1] << 7) | (_bytes[2] << 14)
    elif len(_bytes) == 4:
        _val = _bytes[0] | (_bytes[1] << 7) | (_bytes[2] << 14) | (_bytes[3] << 21)
    else:
        raise ValueError, "unexpected byte array length: %d" % len(_bytes)
    return _offset, (_fac * _val)

def get_text_string(data, offset):
    _bitpos, _len = get_bit_short(data, offset)
    _bitlen = _len * 8
    _string = get_bits(data, _bitlen, _bitpos)
    _bitpos = _bitpos + _bitlen
    return _bitpos, _string

def get_handle(data, offset):
    _code = get_bits(data, 4, offset)
    _counter = get_bits(data, 4, (offset + 4))
    _read = 8
    _hlist = []
    if _counter:
        _hlen = _counter * 8
        _handle = get_bits(data, _hlen, (offset + 8))
        _read = _read + _hlen
        if _hlen > 8: # convert string into list of bytes
            for _chr in _handle:
                _hlist.append(ord(_chr))
        else:
            _hlist.append(_handle)
    return (offset + _read), (_code, _counter) + tuple(_hlist)

def get_modular_short(handle):
    _shorts = []
    _short = struct.unpack('>h', handle.read(2))[0] # msb first
    while (_short & 0x80): # test high bit in lsb byte
        _shorts.append(_short)
        _short = struct.unpack('>h', handle.read(2))[0] # msb first
    _shorts.append(_short)
    for _i in range(len(_shorts)): # reverse bytes in shorts
        _short = _shorts[_i]
        _shorts[_i] = ((_short & 0xff00) >> 8) | ((_short & 0xff) << 8)
    _slen = len(_shorts)
    if _slen == 1:
        _size = _shorts[0] & 0x7fff
    elif _slen == 2:
        _tmp = _shorts[0]
        _shorts[0] = _shorts[1]
        _shorts[1] = _tmp
        _size = ((_shorts[0] & 0x7fff) << 15) | (_shorts[1] & 0x7fff)
    else:
        raise ValueError, "Unexpected array length: %d" % _slen
    return _size

#
# mask1: bit mask to apply to the current byte
# lshift: left shift amount of mask results
# mask2: bit mask to apply to the next byte
# rshift: right shift amount of the mask results
#
_mask_table = [
    (0xff, 0, 0x00, 0), # bit offset == 0
    (0x7f, 1, 0x80, 7), # bit offset == 1
    (0x3f, 2, 0xc0, 6), # bit offset == 2
    (0x1f, 3, 0xe0, 5), # bit offset == 3
    (0x0f, 4, 0xf0, 4), # bit offset == 4
    (0x07, 5, 0xf8, 3), # bit offset == 5
    (0x03, 6, 0xfc, 2), # bit offset == 6
    (0x01, 7, 0xfe, 1), # bit offset == 7
    ]

def get_bits(data, count, offset):
    # dbg_print("debugging on")
    dbg_print("passed %d data length with %d count at %d offset" %
              (len(data), count, offset))
    _idx = offset / 8 # index to the byte offset
    _bitidx = offset % 8 # index to the bit offset within the byte
    _mask1, _lsh, _mask2, _rsh = _mask_table[_bitidx]
    _binc = 8 - _bitidx # bits available in current byte
    _read = 0
    _rem = count
    _byte = 0x0
    _bytes = []
    while _read < count:
        if _rem > _binc: # need more bits than this byte can provide
            dbg_print("_rem > _binc")
            _b1 = (data[_idx] & _mask1)
            _read = _read + _binc
            if not isinstance(_rem, int):
                dbg_print("rem type: " + str(type(_rem)))
                dbg_print("rem: " + str(_rem))
            if not isinstance(_binc, int):
                dbg_print("binc type: " + str(type(_binc)))
                dbg_print("binc: " + str(_binc))
            _rem = _rem - _binc
        else: # this byte can give all the bits needed
            dbg_print("_rem <= _binc")
            _byte = _b1 = ((data[_idx] & _mask1) >> (8 - _bitidx - _rem))
            _read = _read + _rem
            _rem = 0
        if _read < count: # need bits from next byte
            dbg_print("_read %d < %d count" % (_read, count))
            _idx = _idx + 1
            if _rem > _bitidx: # use all bitidx bits - make a complete byte
                dbg_print("_rem (%d) > (%d) _bitidx" % (_rem, _bitidx))
                dbg_print("index %d of %d" % (_idx, len(data)))
                _b2 = (data[_idx] & _mask2)
                _byte = (_b1 << _lsh) | (_b2 >> _rsh)
                _read = _read + _bitidx
                _rem = _rem - _bitidx
            else: # use some bitidx to complete bit count request
                dbg_print("_rem <= _bitidx")
                _mask = _mask_table[_rem][2] # mask for current byte
                _b2 = data[_idx] & _mask
                _byte = (_b1 << _rem) | (_b2 >> (8 - _rem))
                _read = _read + _rem
                _rem = 0
        if count > 8:
            _bytes.append(chr(_byte))
    if len(_bytes):
        return "".join(_bytes)
    return _byte

def test_bit(data, offset):
    _idx = offset / 8 # index to the byte offset
    _bitidx = offset % 8 # index to the bit offset within the byte
    _mask = 0x1 << (7 - _bitidx)
    _val = False
    if (data[_idx] & _mask):
        _val = True
    return (offset + 1), _val

#
# debug routines
#

def set_nodebug():
    import dwgutil
    dwgutil._debug = 0

def set_debug(filename=None):
    import dwgutil
    dwgutil._debug = 1
    if filename != None:
        dwgutil.dbg_handle = open(filename, 'w')
    else:
        dwgutil.dbg_handle = sys.stdout

def dbg_print(*s):
    if debug:
        string = ""
        for arg in s:
            string += str(arg) + " "
        string += "\n"
        dbg_handle.write(string)



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Deleting objects
#

import Generic.point
import Generic.circle
import Generic.arc
import Generic.conobject
import Generic.hcline
import Generic.vcline
import Generic.acline
import Generic.segjoint
import Generic.cline
import Generic.ccircle
import Generic.leader
import Generic.polyline
import Generic.text
import Generic.dimension

def delete_objects(layer, objlist):
    """Remove a list of objects from a Layer.

delete_objects(objlist)

The objlist argument must be a list containing objects stored
within the layer.
    """
    if not isinstance(objlist, list):
        raise TypeError, "Invalid object list: " + `objlist`
    for _obj in objlist:
        _lobj = layer.findObject(_obj)
        if _lobj is not _obj:
            raise ValueError, "Object not found in layer: " + `_obj`
    _objdict = {}
    for _obj in objlist:
        _objdict[id(_obj)] = True
    for _obj in objlist:
        if isinstance(_obj, Generic.point.Point):
            continue
        elif isinstance(_obj, Generic.segment.Segment):
            _p1, _p2 = _obj.getEndpoints()
            _pid = id(_p1)
            if _pid in _objdict:
                _objdict[_pid] = False
            _pid = id(_p2)
            if _pid in _objdict:
                _objdict[_pid] = False
        elif isinstance(_obj, Generic.arc.Arc):
            _cp = _obj.getCenter()
            _pid = id(_cp)
            if _pid in _objdict:
                _objdict[_pid] = False
            for _ep in _obj.getEndpoints():
                _lp = layer.find('point', _ep[0], _ep[1])
                assert _lp is not None, "Missing arc endpoint:" + str(_ep)
                _pid = id(_lp)
                if _pid in _objdict:
                    _objdict[_pid] = False
        elif isinstance(_obj, (Generic.circle.Circle,
                               Generic.ccircle.CCircle)):
            _cp = _obj.getCenter()
            _pid = id(_cp)
            if _pid in _objdict:
                _objdict[_pid] = False
        elif isinstance(_obj, (Generic.hcline.HCLine,
                               Generic.vcline.VCLine,
                               Generic.acline.ACLine)):
            _lp = _obj.getLocation()
            _pid = id(_lp)
            if _pid in _objdict:
                _objdict[_pid] = False
        elif isinstance(_obj, Generic.cline.CLine):
            _p1, _p2 = _obj.getKeypoints()
            _pid = id(_p1)
            if _pid in _objdict:
                _objdict[_pid] = False
            _pid = id(_p2)
            if _pid in _objdict:
                _objdict[_pid] = False
        elif isinstance(_obj, (Generic.segjoint.Chamfer,
                               Generic.segjoint.Fillet)):
            continue # chamfers/fillets do not delete attached segments
        elif isinstance(_obj, (Generic.leader.Leader,
                               Generic.polyline.Polyline)):
            for _pt in _obj.getPoints():
                _pid = id(_pt)
                if _pid in _objdict:
                    _objdict[_pid] = False
        elif isinstance(_obj, Generic.text.TextBlock):
            continue
        else:
            pass
    for _obj in objlist:
        if _objdict[id(_obj)]:
            layer.delObject(_obj)



#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# functions for handling tangent circles on multiple objects
#

import math

import Generic.hcline
import Generic.vcline
import Generic.acline
import Generic.cline
import Generic.ccircle

#
# common constants
#

_dtr = math.pi/180.0
_piover2 = math.pi/2.0

#
# find the projection point of point (x, y) on a line
# from (x1, y1) to (x2, y2)
#

def _get_two_point_projection(x1, y1, x2, y2, x, y):
    _sqlen = pow((x2 - x1), 2) + pow((y2 - y1), 2)
    _rn = ((x - x1) * (x2 - x1)) + ((y - y1) * (y2 - y1))
    _r = _rn/_sqlen
    _px = x1 + _r * (x2 - x1)
    _py = y1 + _r * (y2 - y1)
    return _px, _py

#
# find the projection point of point (x, y) on a line
# defined by an angle and y intercept
#

def _get_angled_projection(angle, yint, x, y):
    _x1 = 0.0
    _y1 = yint
    _x2 = _x1 + math.cos(angle)
    _y2 = _y1 + math.sin(angle)
    return _get_two_point_projection(_x1, _y1, _x2, _y2, x, y)

def _two_line_tangent(x1, y1, x2, y2, x3, y3, x4, y4, x, y):
    #
    # this function calculates the apprpriate tangent circle
    # for two lines (x1, y1)->(x2, y2) and (x3, y3)->(x4, y4)
    # using a point (x, y) to determine which tangent circle
    # should be defined
    #
    # test if lines are parallel
    #
    _cx = _cy = _radius = None
    _denom = ((x2 - x1) * (y4 - y3)) - ((y2 - y1) * (x4 - x3))
    # print "denom: %g" % _denom
    if abs(_denom) < 1e-10:
        # print "parallel ..."
        if abs(x2 - x1) < 1e-10: # both vertical
            if x < min(x1, x3) or x > max(x1, x3):
                return None
            _cx = (x3 + x1)/2.0
            _cy = y
            _radius = abs(x3 - x1)/2.0
        elif abs(y2 - y1) < 1e-10: # both horizontal
            if y < min(y1, y3) or y > max(y1, y3):
                return None
            _cx = x
            _cy = (y3 + y1)/2.0
            _radius = abs(y3 - y1)/2.0
        else: # both at equal angles
            _ax1, _ay1 = _get_two_point_projection(x1, y1, x2, y2, x, y)
            # print "ax1: %g; ay1: %g" % (_ax1, _ay1)
            _ax2, _ay2 = _get_two_point_projection(x3, y3, x4, y4, x, y)
            # print "ax2: %g; ay2: %g" % (_ax2, _ay2)
            if (x < min(_ax1, _ax2) or
                x > max(_ax1, _ax2) or
                y < min(_ay1, _ay2) or
                y > max(_ay1, _ay2)):
                return None
            _cx = (_ax1 + _ax2)/2.0
            _cy = (_ay1 + _ay2)/2.0
            _radius = math.hypot((_ax1 - _cx), (_ay1 - _cy))
        return _cx, _cy, _radius
    #
    # the lines are not parallel, so we have to test for the
    # different combinations of horizontal/vertical/sloped lines ...
    #
    if abs(y2 - y1) < 1e-10: # horizontal line 1
        # print "horizontal line 1 ..."
        if abs(y4 - y3) < 1e-10: # this should be handled above ...
            # print "horizontal line 2 ..."
            if y < min(y1, y3) or y > max(y1, y3):
                return None
            _cx = x
            _cy = (y1 + y3)/2.0
            _radius = abs(_cy - y1)
        elif abs(x4 - x3) < 1e-10: # vertical line 2
            # print "vertical line 2 ..."
            _a1 = math.pi/4.0
            _a2 = -_a1
        else:
            _angle = math.atan2((y4 - y3), (x4 - x3))
            _a1 = _angle/2.0
            if _a1 > 0.0:
                _a2 = _a1 - _piover2
            else:
                _a2 = _a1 + _piover2
    elif abs(x2 - x1) < 1e-10: # vertical line 1
        # print "vertical line 1 ..."
        if abs(y4 - y3) < 1e-10: # horizontal line2
            # print "horizontal line 2 ..."
            _a1 = math.pi/4.0
            _a2 = -_a1
        elif abs(x4 - x3) < 1e-10: # this should be handled above ...
            if x < min(x1, x3) or x > max(x1, x3):
                return None
            _cx = (x1 + x3)/2.0
            _cy = y
            _radius = abs(_cx - x1)
        else:
            _angle = math.atan2((y4 - y3), (x4 - x3))
            if _angle > 0.0:
                _a1 = (_angle + _piover2)/2.0
                _a2 = _a1 - _piover2
            else:
                _a1 = (_angle - _piover2)/2.0
                _a2 = _a1 + _piover2
    else:
        _angle1 = math.atan2((y2 - y1), (x2 - x1))
        if abs(y4 - y3) < 1e-10: # horizontal line2
            _a1 = _angle1/2.0
            if _a1 > 0.0:
                _a2 = _a1 - _piover2
            else:
                _a2 = _a1 + _piover2
        elif abs(x4 - x3) < 1e-10: # vertical line2
            if _angle1 > 0.0:
                _a1 = (_angle1 + _piover2)/2.0
                _a2 = _a1 - _piover2
            else:
                _a1 = (_angle1 - _piover2)/2.0
                _a2 = _a1 + _piover2
        else:
            _angle2 = math.atan2((y4 - y3), (x4 - x3))
            _a1 = (_angle1 + _angle2)/2.0
            if _a1 > 0.0:
                _a2 = _a1 - _piover2
            else:
                _a2 = _a1 + _piover2
    if _cx is not None and _cy is not None and _radius is not None:
        return _cx, _cy, _radius
    #
    # handle the general case of two lines at arbitrary angles
    #
    # print "arbitrary angles ..."
    # print "a1: %g" % (_a1/_dtr)
    # print "a2: %g" % (_a2/_dtr)
    _rn = ((y1 - y3) * (x4 - x3)) - ((x1 - x3) * (y4 - y3))
    # print "rn: %g" % _rn
    _r = _rn/_denom
    # print "r: %g" % _r
    _ix = x1 + _r * (x2 - x1)
    _iy = y1 + _r * (y2 - y1)
    # print "ix: %g; iy: %g" % (_ix, _iy)
    _m1 = math.tan(_a1)
    _b1 = _iy - (_m1 * _ix)
    # print "line1: m: %g; b: %g" % (_m1, _b1)
    _m2 = math.tan(_a2)
    _b2 = _iy - (_m2 * _ix)
    # print "line2: m: %g; b: %g" %  (_m2, _b2)
    _px1, _py1 = _get_angled_projection(_a1, _b1, x, y)
    # print "px1: %g; py1: %g" % (_px1, _py1)
    _sep1 = math.hypot((_px1 - x), (_py1 - y))
    _px2, _py2 = _get_angled_projection(_a2, _b2, x, y)
    # print "px2: %g; py2: %g" % (_px2, _py2)
    _sep2 = math.hypot((_px2 - x), (_py2 - y))
    if _sep1 < _sep2:
        _cx = _px1
        _cy = _py1
    else:
        _cx = _px2
        _cy = _py2
    _px, _py = _get_two_point_projection(x1, y1, x2, y2, _cx, _cy)
    _radius = math.hypot((_px - _cx), (_py - _cy))
    return _cx, _cy, _radius

#
# horizontal construction line tangents
#

def _hcl_hcl_tangent(hcla, hclb, x, y):
    _hx1, _hy1 = hcla.getLocation().getCoords()
    _hx2, _hy2 = hclb.getLocation().getCoords()
    if y < min(_hy1, _hy2) or y > max(_hy1, _hy2):
        return None
    _cx = x
    _cy = (_hy1 + _hy2)/2.0
    _radius = abs(_cy - _hy1)
    return _cx, _cy, _radius

def _hcl_vcl_tangent(hcl, vcl, x, y):
    _hx, _hy = hcl.getLocation().getCoords()
    _hx2 = _hx + 1.0
    _vx, _vy = vcl.getLocation().getCoords()
    _vy2 = _vy + 1.0
    return _two_line_tangent(_hx, _hy, _hx2, _hy, _vx, _vy, _vx, _vy2, x, y)

def _hcl_acl_tangent(hcl, acl, x, y):
    _hx, _hy = hcl.getLocation().getCoords()
    _hx2 = _hx + 1.0
    _ax, _ay = acl.getLocation().getCoords()
    _angle = acl.getAngle() * _dtr
    _ax2 = _ax + math.cos(_angle)
    _ay2 = _ay + math.sin(_angle)
    return _two_line_tangent(_hx, _hy, _hx2, _hy, _ax, _ay, _ax2, _ay2, x, y)

def _hcl_cl_tangent(hcl, cl, x, y):
    _hx, _hy = hcl.getLocation().getCoords()
    _hx2 = _hx + 1.0
    _p1, _p2 = cl.getKeypoints()
    _x1, _y1 = _p1.getCoords()
    _x2, _y2 = _p2.getCoords()
    return _two_line_tangent(_hx, _hy, _hx2, _hy, _x1, _y1, _x2, _y2, x, y)

#
# the HCLine-CCircle tangent circle problem has been generalized
# to solve the {VCLine/ACLine/CLine}-CCircle tangent calculation
#

def _gen_cline_ccircle_tangent(radius, hy, x):
    #
    # center of ccircle : (0,0)
    # radius of ccircle : r
    # distance between ccircle and hcl: hy
    # x-coordinate of mouse: x
    #
    # center of tangent circle : (px, py)
    #
    # projection point on hcl: (px, hy)
    #
    # Projection point outside the radius of the circle gives:
    #
    # math.hypot((px - cx), (py - cy)) == math.hypot((px - px), (py - hy)) + r
    #
    # Projection point inside the radius of the circle gives:
    #
    # math.hypot((px - cx), (py - cy)) + rt = r
    #
    # rt == radius of tangent circle
    #
    # Distance from projection point to tangent circle center
    #
    # math.hypot((px - px), (py - hy)) = rt
    #
    # lots of algebra reduces to the following two cases
    #
    _cx = x
    if hy > 0.0:
        _num = pow(x, 2) - pow(hy, 2) - (2.0 * hy * radius) - pow(radius, 2)
        _den = (-2.0 * hy) - (2.0 * radius)
    else:
        _num = pow(x, 2) - pow(hy, 2) + (2.0 * hy * radius) - pow(radius, 2)
        _den = (2.0 * radius) - (2.0 * hy)
    # print "num: %g" % _num
    # print "den: %g" % _den
    _cy = _num/_den
    _radius = abs(_cy - hy)
    return _cx, _cy, _radius

def _hcl_cc_tangent(hcl, cc, x, y):
    _hx, _hy = hcl.getLocation().getCoords()
    # print "hy: %g" % _hy
    _ccx, _ccy = cc.getCenter().getCoords()
    _rad = cc.getRadius()
    #
    # transform the coords into system where circle center is (0,0)
    #
    _sep = _hy - _ccy
    _xproj = x - _ccx
    _tcx, _tcy, _tcrad = _gen_cline_ccircle_tangent(_rad, _sep, _xproj)
    #
    # transform result back into real coordinates
    #
    _cx = _tcx + _ccx
    _cy = _tcy + _ccy
    return _cx, _cy, _tcrad

#
# vertical construction line
#

def _vcl_vcl_tangent(vcl1, vcl2, x, y):
    _vx1, _vy1 = vcl1.getLocation().getCoords()
    _vx2, _vy2 = vcl2.getLocation().getCoords()
    if x < min(_vx1, _vx2) or x > max(_vx1, _vx2):
        return None
    _cx = (_vx1 + _vx2)/2.0
    _cy = y
    _radius = abs(_cx - _vx1)
    return _cx, _cy, _radius

def _vcl_acl_tangent(vcl, acl, x, y):
    _vx, _vy = vcl.getLocation().getCoords()
    _vy2 = _vy + 1.0
    _ax, _ay = acl.getLocation().getCoords()
    _angle = acl.getAngle() * _dtr
    _ax2 = _ax + math.cos(_angle)
    _ay2 = _ay + math.sin(_angle)
    return _two_line_tangent(_vx, _vy, _vx, _vy2, _ax, _ay, _ax2, _ay2, x, y)

def _vcl_cl_tangent(hcl, cl, x, y):
    _vx, _vy = hcl.getLocation().getCoords()
    _vy2 = _vy + 1.0
    _p1, _p2 = cl.getKeypoints()
    _x1, _y1 = _p1.getCoords()
    _x2, _y2 = _p2.getCoords()
    return _two_line_tangent(_vx, _vy, _vx, _vy2, _x1, _y1, _x2, _y2, x, y)

def _vcl_cc_tangent(vcl, cc, x, y):
    _vx, _vy = vcl.getLocation().getCoords()
    # print "vx: %g" % _vx
    _ccx, _ccy = cc.getCenter().getCoords()
    _rad = cc.getRadius()
    #
    # transform the coords into system where circle center is (0,0) and
    # rotate 90 degrees
    #
    _sep = _vx - _ccx
    _xproj = y - _ccy
    _tcx, _tcy, _tcrad = _gen_cline_ccircle_tangent(_rad, _sep, _xproj)
    #
    # transform result back into real coordinates
    #
    _cx = _tcy + _ccx
    _cy = _tcx + _ccy
    return _cx, _cy, _tcrad

#
# angular construction line
#

def _acl_acl_tangent(acl1, acl2, x, y):
    _ax1, _ay1 = acl1.getLocation().getCoords()
    _angle1 = acl1.getAngle() * _dtr
    _ax2 = _ax1 + math.cos(_angle1)
    _ay2 = _ay1 + math.sin(_angle1)
    _ax3, _ay3 = acl2.getLocation().getCoords()
    _angle2 = acl2.getAngle() * _dtr
    _ax4 = _ax3 + math.cos(_angle2)
    _ay4 = _ay3 + math.sin(_angle2)
    return _two_line_tangent(_ax1, _ay1, _ax2, _ay2, _ax3, _ay3, _ax4, _ay4,
                             x, y)

def _acl_cl_tangent(acl, cl, x, y):
    _ax1, _ay1 = acl.getLocation().getCoords()
    _angle = acl.getAngle() * _dtr
    _ax2 = _ax1 + math.cos(_angle)
    _ay2 = _ay1 + math.sin(_angle)
    _p1, _p2 = cl.getKeypoints()
    _x1, _y1 = _p1.getCoords()
    _x2, _y2 = _p2.getCoords()
    return _two_line_tangent(_ax1, _ay1, _ax2, _ay2, _x1, _y1, _x2, _y2, x, y)

def _acl_cc_tangent(acl, cc, x, y):
    _ax, _ay = acl.getLocation().getCoords()
    # print "ax: %g; ay: %g" % (_ax, _ay)
    _angle = acl.getAngle()
    _ccx, _ccy = cc.getCenter().getCoords()
    _rad = cc.getRadius()
    #
    # transform the coords into the system where the circle center is (0,0)
    # and rotate so the ACLine is horizontal
    #
    _apx, _apy = acl.getProjection(_ccx, _ccy)
    _sep = math.hypot((_apx - _ccx), (_apy - _ccy))
    if abs(_angle) < 1e-10: # horizontal
        _sine = 0.0
        if _apy > _ccy: # system rotated 0.0
            _cosine = 1.0
        else: # system rotated 180.0
            _cosine = -1.0
    elif abs(abs(_angle) - 90.0) < 1e-10: # vertical
        _cosine = 0.0
        if _apx > _ccx: # system rotated 90.0
            _sine = 1.0
        else: # system rotated -90.0
            _sine = -1.0
    else:
        _angle = _piover2 - math.atan2((_apy - _ccy), (_apx - _ccx))
        _sine = math.sin(_angle)
        _cosine = math.cos(_angle)
    #
    # transform (x, y)
    #
    _tx1 = x - _ccx
    _ty1 = y - _ccy
    #
    # transform by rotating through _negative_ angle to
    # map to horizontal line
    #
    _tx = (_tx1 * _cosine) - (_ty1 * _sine)
    _ty = (_tx1 * _sine) + (_ty1 * _cosine)
    _tcx, _tcy, _tcrad = _gen_cline_ccircle_tangent(_rad, _sep, _tx)
    #
    # transform result back into real coordinates
    #
    _cx = ((_tcx * _cosine) + (_tcy * _sine)) + _ccx
    _cy = (-(_tcx * _sine) + (_tcy * _cosine)) + _ccy
    return _cx, _cy, _tcrad

#
# two-point construction line
#

def _cl_cl_tangent(cl1, cl2, x, y):
    _p1, _p2 = cl1.getKeypoints()
    _x1, _y1 = _p1.getCoords()
    _x2, _y2 = _p2.getCoords()
    _p3, _p4 = cl2.getKeypoints()
    _x3, _y3 = _p3.getCoords()
    _x4, _y4 = _p4.getCoords()
    return _two_line_tangent(_x1, _y1, _x2, _y2, _x3, _y3, _x4, _y4, x, y)

def _cl_cc_tangent(cl, cc, x, y):
    _p1, _p2 = cl.getKeypoints()
    _x1, _y1 = _p1.getCoords()
    # print "x1: %g; y1: %g" % (_x1, _y1)
    _x2, _y2 = _p2.getCoords()
    # print "x2: %g; y2: %g" % (_x2, _y2)
    _ccx, _ccy = cc.getCenter().getCoords()
    # print "ccx: %g; ccy: %g" % (_ccx, _ccy)
    _rad = cc.getRadius()
    #
    # transform the coords into the system where the circle center is (0,0)
    # and rotate so the CLine is horizontal
    #
    _apx, _apy = cl.getProjection(_ccx, _ccy)
    # print "apx: %g; apy: %g" % (_apx, _apy)
    _sep = math.hypot((_apx - _ccx), (_apy - _ccy))
    # print "sep: %g" % _sep
    #
    # use the line (ccx, ccy) to (apx, apy) to determine the
    # angular rotation
    #
    if abs(_apx - _ccx) < 1e-10: # cline is horizontal
        _sine = 0.0
        if _apy > _ccy: # system rotated 0.0
            _cosine = 1.0
        else: # system rotated 180.0
            _cosine = -1.0
    elif abs(_apy - _ccy) < 1e-10: # cline is vertical; system rotated -90.0
        _cosine = 0.0
        if _apx > _ccx: # system rotated 90.0
            _sine = 1.0
        else: # system rotated -90.0
            _sine = -1.0
    else:
        _angle = _piover2 - math.atan2((_apy - _ccy), (_apx - _ccx))
        # print "angle: %g" % _angle
        _sine = math.sin(_angle)
        _cosine = math.cos(_angle)
    # print "sin(angle): %g" % _sine
    # print "cos(angle): %g" % _cosine
    #
    # transform (x, y) into
    _tx1 = x - _ccx
    _ty1 = y - _ccy
    # print "tx1: %g; ty1: %g" % (_tx1, _ty1)
    #
    # transform by rotating through angle to
    # map to horizontal line
    #
    _tx = (_tx1 * _cosine) - (_ty1 * _sine)
    _ty = (_tx1 * _sine) + (_ty1 * _cosine)
    # print "tx: %g; ty: %g" % (_tx, _ty)
    _tcx, _tcy, _tcrad = _gen_cline_ccircle_tangent(_rad, _sep, _tx)
    #
    # transform result back into real coordinates
    #
    # print "tcx: %g: tcy: %g" % (_tcx, _tcy)
    _cx = ((_tcx * _cosine) + (_tcy * _sine)) + _ccx
    _cy = (-(_tcx * _sine) + (_tcy * _cosine)) + _ccy
    # print "cx: %g; cy %g" % (_cx, _cy)
    return _cx, _cy, _tcrad

def calc_tangent_circle(obja, objb, x, y):
    _x = x
    if not isinstance(_x, float):
        _x = float(x)
    _y = y
    if not isinstance(_y, float):
        _y = float(y)
    _tandata = None
    if isinstance(obja, Generic.hcline.HCLine):
        if isinstance(objb, Generic.hcline.HCLine):
            _tandata = _hcl_hcl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.vcline.VCLine):
            _tandata = _hcl_vcl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.acline.ACLine):
            _tandata = _hcl_acl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.cline.CLine):
            _tandata = _hcl_cl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.ccircle.CCircle):
            _tandata = _hcl_cc_tangent(obja, objb, _x, _y)
    elif isinstance(obja, Generic.vcline.VCLine):
        if isinstance(objb, Generic.hcline.HCLine):
            _tandata = _hcl_vcl_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.vcline.VCLine):
            _tandata = _vcl_vcl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.acline.ACLine):
            _tandata = _vcl_acl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.cline.CLine):
            _tandata = _vcl_cl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.ccircle.CCircle):
            _tandata = _vcl_cc_tangent(obja, objb, _x, _y)
    elif isinstance(obja, Generic.acline.ACLine):
        if isinstance(objb, Generic.hcline.HCLine):
            _tandata = _hcl_acl_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.vcline.VCLine):
            _tandata = _vcl_acl_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.acline.ACLine):
            _tandata = _acl_acl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.cline.CLine):
            _tandata = _acl_cl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.ccircle.CCircle):
            _tandata = _acl_cc_tangent(obja, objb, _x, _y)
    elif isinstance(obja, Generic.cline.CLine):
        if isinstance(objb, Generic.hcline.HCLine):
            _tandata = _hcl_cl_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.vcline.VCLine):
            _tandata = _vcl_cl_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.acline.ACLine):
            _tandata = _acl_cl_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.cline.CLine):
            _tandata = _cl_cl_tangent(obja, objb, _x, _y)
        elif isinstance(objb, Generic.ccircle.CCircle):
            _cl_cc_tangent(obja, objb, _x, _y)
    elif isinstance(obja, Generic.ccircle.CCircle):
        if isinstance(objb, Generic.hcline.HCLine):
            _tandata = _hcl_cc_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.vcline.VCLine):
            _tandata = _vcl_cc_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.acline.ACLine):
            _tandata = _acl_cc_tangent(objb, obja, _x, _y)
        elif isinstance(objb, Generic.cline.CLine):
            _tandata = _cl_cc_tangent(objb, obja, _x, _y)
        else:
            pass # CCircle/CCircle tangent circles to do later ...
    return _tandata

#
# calculate the possible tangent lines between two circles
#

def _calc_values(ax, ay, bx, by, cx, cy):
    """This function was used for debugging"""
    _den = pow((bx - ax), 2) + pow((by - ay), 2)
    _num = ((cx - ax) * (bx - ax)) + ((cy - ay) * (by - ay))
    _r = _num/_den
    # print "r: %g" % _r
    _num = ((ay - cy) * (bx - ax)) - ((ax - cx) * (by - ay))
    _s = _num/_den
    # print "s: %g" % _s
    _sep = abs(_s) * math.sqrt(_den)
    # print "sep: %g" % _sep
    # return _r, _s, _sep

def _calc_tangent_triangle(r1, r2, sep, ip):
    _sine = r1/abs(ip)
    # print "sin: %g" % _sine
    _angle = math.asin(_sine)
    # print "angle: %g" % (_angle * (180.0/math.pi))
    _tan = math.tan(_angle)
    # print "tan(angle): %g" % _tan
    _cosine = math.cos(_angle)
    # print "cos(angle): %g" % _cosine
    _tanlen = r1/_tan
    # print "tanlen: %g" % _tanlen
    if ip < 0.0: # r1 < r2 and intersection point left of r1
        assert abs(ip) > r1, "Expected ip beyond radius: %g < %g" % (ip, r1)
        _tx1 =  ip + (_tanlen * _cosine)
    else:
        _tx1 =  ip - (_tanlen * _cosine)
    # print "tx1: %g" % _tx1
    _ty1 = _tanlen * _sine
    # print "ty1: %g" % _ty1
    _dist = math.hypot(_tx1, _ty1)
    # print "dist: %g" % _dist
    assert abs(_dist - r1) < 1e-10, "Invalid tangent point for circle 1"
    _tanlen = r2/_tan
    # print "tanlen: %g" % _tanlen
    if ip < 0.0: # see above
        _tx2 = ip + (_tanlen * _cosine)
        _ty2 = _tanlen * _sine
    elif ip > (sep + r2): # only possible if r1 > r2
        _tx2 = ip - (_tanlen * _cosine)
        _ty2 = _tanlen * _sine
    else:
        _tx2 = ip + (_tanlen * _cosine)
        _ty2 = -1.0 * _tanlen * _sine
    # print "tx2: %g" % _tx2
    # print "ty2: %g" % _ty2
    _dist = math.hypot((_tx2 - sep), _ty2)
    # print "dist: %g" % _dist
    assert abs(_dist - r2) < 1e-10, "Invalid tangent point for circle 2"
    return _tx1, _ty1, _tx2, _ty2

def calc_two_circle_tangents(r1, r2, sep):
    # print "in calc_two_circle_tangents() ..."
    _r1 = r1
    if not isinstance(_r1, float):
        _r1 = float(r1)
    if not _r1 > 0.0:
        raise ValueError, "Invalid radius: %g" % _r1
    _r2 = r2
    if not isinstance(_r2, float):
        _r2 = float(r2)
    if not _r2 > 0.0:
        raise ValueError, "Invalid radius: %g" % _r2
    _sep = sep
    if not isinstance(_sep, float):
        _sep = float(sep)
    _tangents = []
    if (abs(_sep) + min(_r1, _r2)) > max(_r1, _r2): # small circle not within larger
        if abs(_r1 - _r2) < 1e-10:
            # print "same radii ..."
            _tangents.append((0.0, _r1, _sep, _r2))
            _tangents.append((0.0, -_r1, _sep, -_r2))
            if abs(_sep) > _r1 + _r2:
                _mid = _sep/2.0
                _angle = math.asin(_r1/_mid)
                _tanlen = _r1/math.tan(_angle)
                _xt = _tanlen * math.cos(_angle)
                _yt = _tanlen * math.sin(_angle)
                _tx1 = _mid - _xt
                _ty1 = _yt
                _tx2 = _mid + _xt
                _ty2 = -_yt
                _tangents.append((_tx1, _ty1, _tx2, _ty2))
                # _calc_values(_tx1, _ty1, _tx2, _ty2, 0.0, 0.0)
                # _calc_values(_tx1, _ty1, _tx2, _ty2, _sep, 0.0)
                _tangents.append((_tx1, -_ty1, _tx2, -_ty2))
                # _calc_values(_tx1, -_ty1, _tx2, -_ty2, 0.0, 0.0)
                # _calc_values(_tx1, -_ty1, _tx2, -_ty2, _sep, 0.0)
        else:
            _alpha = pow((_r1/_r2), 2)
            # print "alpha: %g" % _alpha
            _a = (1.0 - _alpha)
            # print "a: %g" % _a
            _b = (2.0 * _alpha * _sep)
            # print "b: %g" % _b
            _c = (-1.0 * _alpha * pow(_sep, 2))
            # print "c: %g" % _c
            _det = pow(_b, 2) - (4.0 * _a * _c)
            # print "det: %g" % _det
            if _det > 0.0: # can this ever be negative?
                # print "r1: %g" % _r1
                # print "r2: %g" % _r2
                # print "sep: %g" % _sep
                _denom = 2.0 * _a
                _det_sqrt = math.sqrt(_det)
                _num = (-1.0 * _b) + _det_sqrt
                _offset = _num/_denom
                # print "offset: %g" % _offset
                if (_r1 > _r2):
                    # print "r1 > r2"
                    if ((_offset > (_sep + _r2)) or
                        ((_offset > _r1) and (_offset < (_sep - _r2)))):
                        _tpts = _calc_tangent_triangle(_r1, _r2, _sep, _offset)
                        _tangents.append(_tpts)
                        _tx1, _ty1, _tx2, _ty2 = _tpts
                        # _calc_values(_tx1, _ty1, _tx2, _ty2, 0.0, 0.0)
                        # _calc_values(_tx1, _ty1, _tx2, _ty2, _sep, 0.0)
                        # _calc_values(_tx2, _ty2, _tx1, _ty1, _sep, 0.0)
                        # _calc_values(_tx2, _ty2, _tx1, _ty1, 0.0, 0.0)
                        _tpts = (_tx1, - _ty1, _tx2, -_ty2)
                        _tangents.append(_tpts)
                        # _calc_values(_tx1, -_ty1, _tx2,-_ty2, 0.0, 0.0)
                        # _calc_values(_tx1, -_ty1, _tx2, -_ty2, _sep, 0.0)
                        # _calc_values(_tx2, -_ty2, _tx1, -_ty1, _sep, 0.0)
                        # _calc_values(_tx2, -_ty2, _tx1, -_ty1, 0.0, 0.0)
                else: # _r1 < _r2
                   #  print "r1 < r2"
                    if ((_offset < -_r1) or
                        ((_offset > _r1) and (_offset < (_sep - _r2)))):
                        _tpts = _calc_tangent_triangle(_r1, _r2, _sep, _offset)
                        _tangents.append(_tpts)
                        _tx1, _ty1, _tx2, _ty2 = _tpts
                        # _calc_values(_tx1, _ty1, _tx2, _ty2, 0.0, 0.0)
                        # _calc_values(_tx1, _ty1, _tx2, _ty2, _sep, 0.0)
                        # _calc_values(_tx2, _ty2, _tx1, _ty1, _sep, 0.0)
                        # _calc_values(_tx2, _ty2, _tx1, _ty1, 0.0, 0.0)
                        _tpts = (_tx1, - _ty1, _tx2, -_ty2)
                        _tangents.append(_tpts)
                        # _calc_values(_tx1, -_ty1, _tx2,-_ty2, 0.0, 0.0)
                        # _calc_values(_tx1, -_ty1, _tx2, -_ty2, _sep, 0.0)
                        # _calc_values(_tx2, -_ty2, _tx1, -_ty1, _sep, 0.0)
                        # _calc_values(_tx2, -_ty2, _tx1, -_ty1, 0.0, 0.0)
                _num = (-1.0 * _b) - _det_sqrt
                _offset = _num/_denom
                # print "offset: %g" % _offset
                if (_r1 > _r2):
                    # print "r1 > r2"
                    if ((_offset > (_sep + _r2)) or
                        ((_offset > _r1) and (_offset < (_sep - _r2)))):
                        _tpts = _calc_tangent_triangle(_r1, _r2, _sep, _offset)
                        _tangents.append(_tpts)
                        _tx1, _ty1, _tx2, _ty2 = _tpts
                        _tpts = (_tx1, - _ty1, _tx2, -_ty2)
                        _tangents.append(_tpts)
                else: # _r1 < _r2
                    # print "r1 < r2"
                    if ((_offset < -_r1) or
                        ((_offset > _r1) and (_offset < (_sep - _r2)))):
                        _tpts = _calc_tangent_triangle(_r1, _r2, _sep, _offset)
                        _tangents.append(_tpts)
                        _tx1, _ty1, _tx2, _ty2 = _tpts
                        # _calc_values(_tx1, _ty1, _tx2, _ty2, 0.0, 0.0)
                        # _calc_values(_tx1, _ty1, _tx2, _ty2, _sep, 0.0)
                        # _calc_values(_tx2, _ty2, _tx1, _ty1, _sep, 0.0)
                        # _calc_values(_tx2, _ty2, _tx1, _ty1, 0.0, 0.0)
                        _tpts = (_tx1, - _ty1, _tx2, -_ty2)
                        _tangents.append(_tpts)
                        # _calc_values(_tx1, -_ty1, _tx2,-_ty2, 0.0, 0.0)
                        # _calc_values(_tx1, -_ty1, _tx2, -_ty2, _sep, 0.0)
                        # _calc_values(_tx2, -_ty2, _tx1, -_ty1, _sep, 0.0)
                        # _calc_values(_tx2, -_ty2, _tx1, -_ty1, 0.0, 0.0)
    return _tangents
